import pytest

from vyper.exceptions import StaticAssertionException
from vyper.venom.analysis.analysis import IRAnalysesCache
from vyper.venom.basicblock import IRBasicBlock, IRLabel, IRLiteral, IRVariable
from vyper.venom.context import IRContext
from vyper.venom.passes.make_ssa import MakeSSA
from vyper.venom.passes.sccp import SCCP
from vyper.venom.passes.sccp.sccp import LatticeEnum


def test_simple_case():
    ctx = IRContext()
    fn = ctx.create_function("_global")

    bb = fn.get_basic_block()
    p1 = bb.append_instruction("param")
    op1 = bb.append_instruction("store", 32)
    op2 = bb.append_instruction("store", 64)
    op3 = bb.append_instruction("add", op1, op2)
    bb.append_instruction("return", p1, op3)

    ac = IRAnalysesCache(fn)
    MakeSSA(ac, fn).run_pass()
    sccp = SCCP(ac, fn)
    sccp.run_pass()

    assert sccp.lattice[IRVariable("%1")] == LatticeEnum.BOTTOM
    assert sccp.lattice[IRVariable("%2")].value == 32
    assert sccp.lattice[IRVariable("%3")].value == 64
    assert sccp.lattice[IRVariable("%4")].value == 96


def test_branch_eliminator_simple():
    ctx = IRContext()
    fn = ctx.create_function("_global")

    bb1 = fn.get_basic_block()

    br1 = IRBasicBlock(IRLabel("then"), fn)
    br1.append_instruction("stop")
    br2 = IRBasicBlock(IRLabel("else"), fn)
    br2.append_instruction("jmp", IRLabel("foo"))

    fn.append_basic_block(br1)
    fn.append_basic_block(br2)

    bb1.append_instruction("jnz", IRLiteral(1), br1.label, br2.label)

    bb2 = IRBasicBlock(IRLabel("foo"), fn)
    bb2.append_instruction("jnz", IRLiteral(0), br1.label, br2.label)
    fn.append_basic_block(bb2)

    ac = IRAnalysesCache(fn)
    MakeSSA(ac, fn).run_pass()
    sccp = SCCP(ac, fn)
    sccp.run_pass()

    assert bb1.instructions[-1].opcode == "jmp"
    assert bb1.instructions[-1].operands == [br1.label]
    assert bb2.instructions[-1].opcode == "jmp"
    assert bb2.instructions[-1].operands == [br2.label]


def test_assert_elimination():
    ctx = IRContext()
    fn = ctx.create_function("_global")

    bb = fn.get_basic_block()

    bb.append_instruction("assert", IRLiteral(1))
    bb.append_instruction("assert_unreachable", IRLiteral(1))
    bb.append_instruction("stop")

    ac = IRAnalysesCache(fn)
    MakeSSA(ac, fn).run_pass()
    sccp = SCCP(ac, fn)
    sccp.run_pass()

    for inst in bb.instructions[:-1]:
        assert inst.opcode == "nop"


@pytest.mark.parametrize("asserter", ("assert", "assert_unreachable"))
def test_assert_false(asserter):
    ctx = IRContext()
    fn = ctx.create_function("_global")

    bb = fn.get_basic_block()

    bb.append_instruction(asserter, IRLiteral(0))
    bb.append_instruction("stop")

    ac = IRAnalysesCache(fn)
    MakeSSA(ac, fn).run_pass()
    sccp = SCCP(ac, fn)
    with pytest.raises(StaticAssertionException):
        sccp.run_pass()


def test_cont_jump_case():
    ctx = IRContext()
    fn = ctx.create_function("_global")

    bb = fn.get_basic_block()

    br1 = IRBasicBlock(IRLabel("then"), fn)
    fn.append_basic_block(br1)
    br2 = IRBasicBlock(IRLabel("else"), fn)
    fn.append_basic_block(br2)

    p1 = bb.append_instruction("param")
    op1 = bb.append_instruction("store", 32)
    op2 = bb.append_instruction("store", 64)
    op3 = bb.append_instruction("add", op1, op2)
    bb.append_instruction("jnz", op3, br1.label, br2.label)

    br1.append_instruction("add", op3, 10)
    br1.append_instruction("stop")
    br2.append_instruction("add", op3, p1)
    br2.append_instruction("stop")

    ac = IRAnalysesCache(fn)
    MakeSSA(ac, fn).run_pass()
    sccp = SCCP(ac, fn)
    sccp.run_pass()

    assert sccp.lattice[IRVariable("%1")] == LatticeEnum.BOTTOM
    assert sccp.lattice[IRVariable("%2")].value == 32
    assert sccp.lattice[IRVariable("%3")].value == 64
    assert sccp.lattice[IRVariable("%4")].value == 96
    assert sccp.lattice[IRVariable("%5")].value == 106
    assert sccp.lattice.get(IRVariable("%6")) == LatticeEnum.BOTTOM


def test_cont_phi_case():
    ctx = IRContext()
    fn = ctx.create_function("_global")

    bb = fn.get_basic_block()

    br1 = IRBasicBlock(IRLabel("then"), fn)
    fn.append_basic_block(br1)
    br2 = IRBasicBlock(IRLabel("else"), fn)
    fn.append_basic_block(br2)
    join = IRBasicBlock(IRLabel("join"), fn)
    fn.append_basic_block(join)

    p1 = bb.append_instruction("param")
    op1 = bb.append_instruction("store", 32)
    op2 = bb.append_instruction("store", 64)
    op3 = bb.append_instruction("add", op1, op2)
    bb.append_instruction("jnz", op3, br1.label, br2.label)

    op4 = br1.append_instruction("add", op3, 10)
    br1.append_instruction("jmp", join.label)
    br2.append_instruction("add", op3, p1, ret=op4)
    br2.append_instruction("jmp", join.label)

    join.append_instruction("return", op4, p1)

    ac = IRAnalysesCache(fn)
    MakeSSA(ac, fn).run_pass()
    sccp = SCCP(ac, fn)
    sccp.run_pass()

    assert sccp.lattice[IRVariable("%1")] == LatticeEnum.BOTTOM
    assert sccp.lattice[IRVariable("%2")].value == 32
    assert sccp.lattice[IRVariable("%3")].value == 64
    assert sccp.lattice[IRVariable("%4")].value == 96
    assert sccp.lattice[IRVariable("%5", version=1)].value == 106
    assert sccp.lattice[IRVariable("%5", version=2)] == LatticeEnum.BOTTOM
    assert sccp.lattice[IRVariable("%5")].value == 2


def test_cont_phi_const_case():
    ctx = IRContext()
    fn = ctx.create_function("_global")

    bb = fn.get_basic_block()

    br1 = IRBasicBlock(IRLabel("then"), fn)
    fn.append_basic_block(br1)
    br2 = IRBasicBlock(IRLabel("else"), fn)
    fn.append_basic_block(br2)
    join = IRBasicBlock(IRLabel("join"), fn)
    fn.append_basic_block(join)

    p1 = bb.append_instruction("store", 1)
    op1 = bb.append_instruction("store", 32)
    op2 = bb.append_instruction("store", 64)
    op3 = bb.append_instruction("add", op1, op2)
    bb.append_instruction("jnz", op3, br1.label, br2.label)

    op4 = br1.append_instruction("add", op3, 10)
    br1.append_instruction("jmp", join.label)
    br2.append_instruction("add", op3, p1, ret=op4)
    br2.append_instruction("jmp", join.label)

    join.append_instruction("return", op4, p1)

    ac = IRAnalysesCache(fn)
    MakeSSA(ac, fn).run_pass()
    sccp = SCCP(ac, fn)
    sccp.run_pass()

    assert sccp.lattice[IRVariable("%1")].value == 1
    assert sccp.lattice[IRVariable("%2")].value == 32
    assert sccp.lattice[IRVariable("%3")].value == 64
    assert sccp.lattice[IRVariable("%4")].value == 96
    assert sccp.lattice[IRVariable("%5", version=1)].value == 106
    assert sccp.lattice[IRVariable("%5", version=2)].value == 97
    assert sccp.lattice[IRVariable("%5")].value == 2



from vyper.utils import OrderedSet
from vyper.venom.analysis.analysis import IRAnalysis


class DupRequirementsAnalysis(IRAnalysis):
    def analyze(self):
        for bb in self.function.get_basic_blocks():
            last_liveness = bb.out_vars
            for inst in reversed(bb.instructions):
                inst.dup_requirements = OrderedSet()
                ops = inst.get_inputs()
                for op in ops:
                    if op in last_liveness:
                        inst.dup_requirements.add(op)
                last_liveness = inst.liveness



from collections import defaultdict
from typing import Generic, TypeVar

from vyper import ast as vy_ast
from vyper.evm.opcodes import version_check
from vyper.exceptions import CompilerPanic, StorageLayoutException
from vyper.semantics.analysis.base import VarOffset
from vyper.semantics.data_locations import DataLocation
from vyper.typing import StorageLayout


def set_data_positions(
    vyper_module: vy_ast.Module, storage_layout_overrides: StorageLayout = None
) -> StorageLayout:
    """
    Parse the annotated Vyper AST, determine data positions for all variables,
    and annotate the AST nodes with the position data.

    Arguments
    ---------
    vyper_module : vy_ast.Module
        Top-level Vyper AST node that has already been annotated with type data.
    """
    if storage_layout_overrides is not None:
        # extract code layout with no overrides
        code_offsets = _allocate_layout_r(vyper_module, immutables_only=True)["code_layout"]
        storage_slots = set_storage_slots_with_overrides(vyper_module, storage_layout_overrides)
        return {"storage_layout": storage_slots, "code_layout": code_offsets}

    ret = _allocate_layout_r(vyper_module)
    assert isinstance(ret, defaultdict)
    return dict(ret)  # convert back to dict


_T = TypeVar("_T")
_K = TypeVar("_K")


class InsertableOnceDict(Generic[_T, _K], dict[_T, _K]):
    def __setitem__(self, k, v):
        if k in self:
            raise ValueError(f"{k} is already in dict!")
        super().__setitem__(k, v)


# some name that the user cannot assign to a variable
GLOBAL_NONREENTRANT_KEY = "$.nonreentrant_key"


class SimpleAllocator:
    def __init__(self, max_slot: int = 2**256, starting_slot: int = 0):
        # Allocate storage slots from 0
        # note storage is word-addressable, not byte-addressable
        self._starting_slot = starting_slot
        self._slot = starting_slot
        self._max_slot = max_slot

    def allocate_slot(self, n, var_name, node=None):
        ret = self._slot
        if self._slot + n >= self._max_slot:
            raise StorageLayoutException(
                f"Invalid storage slot, tried to allocate"
                f" slots {self._slot} through {self._slot + n}",
                node,
            )
        self._slot += n
        return ret

    def allocate_global_nonreentrancy_slot(self):
        slot = self.allocate_slot(1, GLOBAL_NONREENTRANT_KEY)
        assert slot == self._starting_slot
        return slot


class Allocators:
    storage_allocator: SimpleAllocator
    transient_storage_allocator: SimpleAllocator
    immutables_allocator: SimpleAllocator

    _global_nonreentrancy_key_slot: int

    def __init__(self):
        self.storage_allocator = SimpleAllocator(max_slot=2**256)
        self.transient_storage_allocator = SimpleAllocator(max_slot=2**256)
        self.immutables_allocator = SimpleAllocator(max_slot=0x6000)

    def get_allocator(self, location: DataLocation):
        if location == DataLocation.STORAGE:
            return self.storage_allocator
        if location == DataLocation.TRANSIENT:
            return self.transient_storage_allocator
        if location == DataLocation.CODE:
            return self.immutables_allocator

        raise CompilerPanic("unreachable")  # pragma: nocover

    def allocate_global_nonreentrancy_slot(self):
        location = get_reentrancy_key_location()

        allocator = self.get_allocator(location)
        slot = allocator.allocate_global_nonreentrancy_slot()
        self._global_nonreentrancy_key_slot = slot

    def get_global_nonreentrant_key_slot(self):
        return self._global_nonreentrancy_key_slot


class OverridingStorageAllocator:
    """
    Keep track of which storage slots have been used. If there is a collision of
    storage slots, this will raise an error and fail to compile
    """

    def __init__(self):
        self.occupied_slots: dict[int, str] = {}

    def reserve_slot_range(self, first_slot: int, n_slots: int, var_name: str) -> None:
        """
        Reserves `n_slots` storage slots, starting at slot `first_slot`
        This will raise an error if a storage slot has already been allocated.
        It is responsibility of calling function to ensure first_slot is an int
        """
        list_to_check = [x + first_slot for x in range(n_slots)]
        self._reserve_slots(list_to_check, var_name)

    def _reserve_slots(self, slots: list[int], var_name: str) -> None:
        for slot in slots:
            self._reserve_slot(slot, var_name)

    def _reserve_slot(self, slot: int, var_name: str) -> None:
        if slot < 0 or slot >= 2**256:
            raise StorageLayoutException(
                f"Invalid storage slot for var {var_name}, out of bounds: {slot}"
            )
        if slot in self.occupied_slots:
            collided_var = self.occupied_slots[slot]
            raise StorageLayoutException(
                f"Storage collision! Tried to assign '{var_name}' to slot {slot} but it has "
                f"already been reserved by '{collided_var}'"
            )
        self.occupied_slots[slot] = var_name


def set_storage_slots_with_overrides(
    vyper_module: vy_ast.Module, storage_layout_overrides: StorageLayout
) -> StorageLayout:
    """
    Set storage layout given a layout override file.
    Returns the layout as a dict of variable name -> variable info
    (Doesn't handle modules, or transient storage)
    """
    ret: InsertableOnceDict[str, dict] = InsertableOnceDict()
    reserved_slots = OverridingStorageAllocator()

    # Search through function definitions to find non-reentrant functions
    for node in vyper_module.get_children(vy_ast.FunctionDef):
        type_ = node._metadata["func_type"]

        # Ignore functions without non-reentrant
        if not type_.nonreentrant:
            continue

        variable_name = GLOBAL_NONREENTRANT_KEY

        # re-entrant key was already identified
        if variable_name in ret:
            continue

        # Expect to find this variable within the storage layout override
        if variable_name in storage_layout_overrides:
            reentrant_slot = storage_layout_overrides[variable_name]["slot"]
            # Ensure that this slot has not been used, and prevents other storage variables
            # from using the same slot
            reserved_slots.reserve_slot_range(reentrant_slot, 1, variable_name)

            type_.set_reentrancy_key_position(VarOffset(reentrant_slot))

            ret[variable_name] = {"type": "nonreentrant lock", "slot": reentrant_slot}
        else:
            raise StorageLayoutException(
                f"Could not find storage_slot for {variable_name}. "
                "Have you used the correct storage layout file?",
                node,
            )

    # Iterate through variables
    for node in vyper_module.get_children(vy_ast.VariableDecl):
        # Ignore immutable parameters
        if node.get("annotation.func.id") == "immutable":
            continue

        varinfo = node.target._metadata["varinfo"]

        # Expect to find this variable within the storage layout overrides
        if node.target.id in storage_layout_overrides:
            var_slot = storage_layout_overrides[node.target.id]["slot"]
            storage_length = varinfo.typ.storage_size_in_words
            # Ensure that all required storage slots are reserved, and prevents other variables
            # from using these slots
            reserved_slots.reserve_slot_range(var_slot, storage_length, node.target.id)
            varinfo.set_position(VarOffset(var_slot))

            ret[node.target.id] = {"type": str(varinfo.typ), "slot": var_slot}
        else:
            raise StorageLayoutException(
                f"Could not find storage_slot for {node.target.id}. "
                "Have you used the correct storage layout file?",
                node,
            )

    return ret


def _get_allocatable(vyper_module: vy_ast.Module) -> list[vy_ast.VyperNode]:
    allocable = (vy_ast.InitializesDecl, vy_ast.VariableDecl)
    return [node for node in vyper_module.body if isinstance(node, allocable)]


def get_reentrancy_key_location() -> DataLocation:
    if version_check(begin="cancun"):
        return DataLocation.TRANSIENT
    return DataLocation.STORAGE


_LAYOUT_KEYS = {
    DataLocation.CODE: "code_layout",
    DataLocation.TRANSIENT: "transient_storage_layout",
    DataLocation.STORAGE: "storage_layout",
}


def _allocate_nonreentrant_keys(vyper_module, allocators):
    SLOT = allocators.get_global_nonreentrant_key_slot()

    for node in vyper_module.get_children(vy_ast.FunctionDef):
        type_ = node._metadata["func_type"]
        if not type_.nonreentrant:
            continue

        # a nonreentrant key can appear many times in a module but it
        # only takes one slot. after the first time we see it, do not
        # increment the storage slot.
        type_.set_reentrancy_key_position(VarOffset(SLOT))


def _allocate_layout_r(
    vyper_module: vy_ast.Module, allocators: Allocators = None, immutables_only=False
) -> StorageLayout:
    """
    Parse module-level Vyper AST to calculate the layout of storage variables.
    Returns the layout as a dict of variable name -> variable info
    """
    global_ = False
    if allocators is None:
        global_ = True
        allocators = Allocators()
        # always allocate nonreentrancy slot, so that adding or removing
        # reentrancy protection from a contract does not change its layout
        allocators.allocate_global_nonreentrancy_slot()

    ret: defaultdict[str, InsertableOnceDict[str, dict]] = defaultdict(InsertableOnceDict)

    # tag functions with the global nonreentrant key
    if not immutables_only:
        _allocate_nonreentrant_keys(vyper_module, allocators)

        layout_key = _LAYOUT_KEYS[get_reentrancy_key_location()]
        # TODO this could have better typing but leave it untyped until
        # we nail down the format better
        if global_ and GLOBAL_NONREENTRANT_KEY not in ret[layout_key]:
            slot = allocators.get_global_nonreentrant_key_slot()
            ret[layout_key][GLOBAL_NONREENTRANT_KEY] = {"type": "nonreentrant lock", "slot": slot}

    for node in _get_allocatable(vyper_module):
        if isinstance(node, vy_ast.InitializesDecl):
            module_info = node._metadata["initializes_info"].module_info
            module_layout = _allocate_layout_r(module_info.module_node, allocators)
            module_alias = module_info.alias
            for layout_key in module_layout.keys():
                assert layout_key in _LAYOUT_KEYS.values()
                ret[layout_key][module_alias] = module_layout[layout_key]
            continue

        assert isinstance(node, vy_ast.VariableDecl)
        # skip non-state variables
        varinfo = node.target._metadata["varinfo"]
        if not varinfo.is_state_variable():
            continue
        location = varinfo.location

        if immutables_only and location != DataLocation.CODE:
            continue

        allocator = allocators.get_allocator(location)
        size = varinfo.get_size()

        # CMC 2021-07-23 note that HashMaps get assigned a slot here
        # using the same allocator (even though there is not really
        # any risk of physical overlap)
        offset = allocator.allocate_slot(size, node.target.id, node)

        varinfo.set_position(VarOffset(offset))

        layout_key = _LAYOUT_KEYS[location]
        type_ = varinfo.typ
        # this could have better typing but leave it untyped until
        # we understand the use case better
        if location == DataLocation.CODE:
            item = {"type": str(type_), "length": size, "offset": offset}
        elif location in (DataLocation.STORAGE, DataLocation.TRANSIENT):
            item = {"type": str(type_), "slot": offset}
        else:  # pragma: nocover
            raise CompilerPanic("unreachable")
        ret[layout_key][node.target.id] = item

    return ret



import pytest

from vyper.exceptions import (
    ArrayIndexException,
    InvalidType,
    StructureException,
    UndeclaredDefinition,
)
from vyper.semantics.data_locations import DataLocation
from vyper.semantics.types import PRIMITIVE_TYPES, HashMapT, SArrayT
from vyper.semantics.types.utils import type_from_annotation

BASE_TYPES = ["int128", "uint256", "bool", "address", "bytes32"]
BYTESTRING_TYPES = ["String", "Bytes"]


@pytest.mark.parametrize("type_str", BASE_TYPES)
@pytest.mark.parametrize("location", iter(DataLocation))
def test_base_types(build_node, type_str, location):
    node = build_node(type_str)
    base_t = PRIMITIVE_TYPES[type_str]

    ann_t = type_from_annotation(node)

    assert base_t == ann_t


@pytest.mark.parametrize("type_str", BYTESTRING_TYPES)
@pytest.mark.parametrize("location", iter(DataLocation))
def test_array_value_types(build_node, type_str, location):
    node = build_node(f"{type_str}[1]")
    base_t = PRIMITIVE_TYPES[type_str](1)

    ann_t = type_from_annotation(node)

    assert base_t == ann_t


@pytest.mark.parametrize("type_str", BASE_TYPES)
@pytest.mark.parametrize("location", iter(DataLocation))
def test_base_types_as_arrays(build_node, type_str, location):
    node = build_node(f"{type_str}[3]")
    base_t = PRIMITIVE_TYPES[type_str]

    ann_t = type_from_annotation(node)

    assert ann_t == SArrayT(base_t, 3)


@pytest.mark.parametrize("type_str", BYTESTRING_TYPES)
@pytest.mark.parametrize("location", iter(DataLocation))
def test_array_value_types_as_arrays(build_node, type_str, location):
    node = build_node(f"{type_str}[1][1]")

    with pytest.raises(StructureException):
        type_from_annotation(node)


@pytest.mark.parametrize("type_str", BASE_TYPES)
@pytest.mark.parametrize("location", iter(DataLocation))
def test_base_types_as_multidimensional_arrays(build_node, namespace, type_str, location):
    node = build_node(f"{type_str}[3][5]")
    base_t = PRIMITIVE_TYPES[type_str]

    ann_t = type_from_annotation(node)

    assert ann_t == SArrayT(SArrayT(base_t, 3), 5)


@pytest.mark.parametrize("type_str", ["int128", "String"])
@pytest.mark.parametrize("idx", ["0", "-1", "0x00", "'1'", "foo", "[1]", "(1,)"])
@pytest.mark.parametrize("location", iter(DataLocation))
def test_invalid_index(build_node, idx, type_str, location):
    node = build_node(f"{type_str}[{idx}]")
    with pytest.raises(
        (ArrayIndexException, InvalidType, StructureException, UndeclaredDefinition)
    ):
        type_from_annotation(node)


@pytest.mark.parametrize("type_str", BASE_TYPES)
@pytest.mark.parametrize("type_str2", BASE_TYPES)
def test_mapping(build_node, type_str, type_str2):
    node = build_node(f"HashMap[{type_str}, {type_str2}]")
    types = PRIMITIVE_TYPES

    ann_t = type_from_annotation(node, DataLocation.STORAGE)

    k_t = types[type_str]
    v_t = types[type_str2]

    assert ann_t == HashMapT(k_t, v_t)


@pytest.mark.parametrize("type_str", BASE_TYPES)
@pytest.mark.parametrize("type_str2", BASE_TYPES)
def test_multidimensional_mapping(build_node, type_str, type_str2):
    node = build_node(f"HashMap[{type_str}, HashMap[{type_str}, {type_str2}]]")
    types = PRIMITIVE_TYPES

    ann_t = type_from_annotation(node, DataLocation.STORAGE)

    k_t = types[type_str]
    v_t = types[type_str2]

    assert ann_t == HashMapT(k_t, HashMapT(k_t, v_t))



from vyper.exceptions import InvalidABIType
from vyper.utils import ceil32


# https://solidity.readthedocs.io/en/latest/abi-spec.html#types
class ABIType:
    # TODO should these methods be properties

    # aka has tail
    def is_dynamic(self):
        raise NotImplementedError("ABIType.is_dynamic")

    # size (in bytes) in the static section (aka 'head')
    # when embedded in a complex type.
    def embedded_static_size(self):
        if self.is_dynamic():
            return 32
        return self.static_size()

    # size bound in the dynamic section (aka 'tail')
    # when embedded in a complex type.
    def embedded_dynamic_size_bound(self):
        if not self.is_dynamic():
            return 0
        return self.size_bound()

    def embedded_min_dynamic_size(self):
        if not self.is_dynamic():
            return 0
        return self.min_size()

    # size (in bytes) of the static section
    def static_size(self):
        raise NotImplementedError("ABIType.static_size")

    # max size (in bytes) in the dynamic section (aka 'tail')
    def dynamic_size_bound(self):
        if not self.is_dynamic():
            return 0
        raise NotImplementedError("ABIType.dynamic_size_bound")

    def size_bound(self):
        return self.static_size() + self.dynamic_size_bound()

    def min_size(self):
        return self.static_size() + self.min_dynamic_size()

    def min_dynamic_size(self):
        if not self.is_dynamic():
            return 0
        raise NotImplementedError("ABIType.min_dynamic_size")

    # The canonical name of the type for calculating the function selector
    def selector_name(self):
        raise NotImplementedError("ABIType.selector_name")

    # Whether the type is a tuple at the ABI level.
    # (This is important because if it does, it needs an offset.
    #   Compare the difference in encoding between `bytes` and `(bytes,)`.)
    def is_complex_type(self):
        raise NotImplementedError("ABIType.is_complex_type")

    def __repr__(self):
        return str({type(self).__name__: vars(self)})


# uint<M>: unsigned integer type of M bits, 0 < M <= 256, M % 8 == 0. e.g. uint32, uint8, uint256.
# int<M>: two’s complement signed integer type of M bits, 0 < M <= 256, M % 8 == 0.
class ABI_GIntM(ABIType):
    def __init__(self, m_bits, signed):
        if not (0 < m_bits <= 256 and 0 == m_bits % 8):
            raise InvalidABIType("Invalid M provided for GIntM")

        self.m_bits = m_bits
        self.signed = signed

    def is_dynamic(self):
        return False

    def static_size(self):
        return 32

    def selector_name(self):
        return ("" if self.signed else "u") + f"int{self.m_bits}"

    def is_complex_type(self):
        return False


# address: equivalent to uint160, except for the assumed interpretation
#   and language typing. For computing the function selector, address is used.
class ABI_Address(ABI_GIntM):
    def __init__(self):
        return super().__init__(160, False)

    def selector_name(self):
        return "address"


# bool: equivalent to uint8 restricted to the values 0 and 1.
#  For computing the function selector, bool is used.
#  (thought: is vyper required to check that the value is restricted to 0 and
#  1, i.e. that 248 bits or 255 bits are zeroed? - CC 20191119)
class ABI_Bool(ABI_GIntM):
    def __init__(self):
        return super().__init__(8, False)

    def selector_name(self):
        return "bool"


# bytes<M>: binary type of M bytes, 0 < M <= 32.
class ABI_BytesM(ABIType):
    def __init__(self, m_bytes):
        if not 0 < m_bytes <= 32:
            raise InvalidABIType("Invalid M for BytesM")

        self.m_bytes = m_bytes

    def is_dynamic(self):
        return False

    def static_size(self):
        return 32

    def selector_name(self):
        return f"bytes{self.m_bytes}"

    def is_complex_type(self):
        return False


# function: an address (20 bytes) followed by a function selector (4 bytes).
# Encoded identical to bytes24.
class ABI_Function(ABI_BytesM):
    def __init__(self):
        return super().__init__(24)

    def selector_name(self):
        return "function"


# <type>[M]: a fixed-length array of M elements, M >= 0, of the given type.
class ABI_StaticArray(ABIType):
    def __init__(self, subtyp, m_elems):
        if not m_elems >= 0:
            raise InvalidABIType("Invalid M")

        self.subtyp = subtyp
        self.m_elems = m_elems

    def is_dynamic(self):
        return self.subtyp.is_dynamic()

    def static_size(self):
        return self.m_elems * self.subtyp.embedded_static_size()

    def dynamic_size_bound(self):
        return self.m_elems * self.subtyp.embedded_dynamic_size_bound()

    def min_dynamic_size(self):
        return self.m_elems * self.subtyp.embedded_min_dynamic_size()

    def selector_name(self):
        return f"{self.subtyp.selector_name()}[{self.m_elems}]"

    def is_complex_type(self):
        return True


class ABI_Bytes(ABIType):
    def __init__(self, bytes_bound):
        if not bytes_bound >= 0:
            raise InvalidABIType("Negative bytes_bound provided to ABI_Bytes")

        self.bytes_bound = bytes_bound

    def is_dynamic(self):
        return True

    # note that static_size for dynamic types is always 0
    # (and embedded_static_size is always 32)
    def static_size(self):
        return 0

    def dynamic_size_bound(self):
        # length word + data
        return 32 + ceil32(self.bytes_bound)

    def min_dynamic_size(self):
        return 32

    def selector_name(self):
        return "bytes"

    def is_complex_type(self):
        return False


class ABI_String(ABI_Bytes):
    def selector_name(self):
        return "string"


class ABI_DynamicArray(ABIType):
    def __init__(self, subtyp, elems_bound):
        if not elems_bound >= 0:
            raise InvalidABIType("Negative bound provided to DynamicArray")

        self.subtyp = subtyp
        self.elems_bound = elems_bound

    def is_dynamic(self):
        return True

    def static_size(self):
        return 0

    def dynamic_size_bound(self):
        subtyp_size = self.subtyp.embedded_static_size() + self.subtyp.embedded_dynamic_size_bound()

        # length + size of embedded children
        return 32 + subtyp_size * self.elems_bound

    def min_dynamic_size(self):
        return 32

    def selector_name(self):
        return f"{self.subtyp.selector_name()}[]"

    def is_complex_type(self):
        return True


class ABI_Tuple(ABIType):
    def __init__(self, subtyps):
        self.subtyps = subtyps

    def is_dynamic(self):
        return any([t.is_dynamic() for t in self.subtyps])

    def static_size(self):
        return sum([t.embedded_static_size() for t in self.subtyps])

    def dynamic_size_bound(self):
        return sum([t.embedded_dynamic_size_bound() for t in self.subtyps])

    def min_dynamic_size(self):
        return sum([t.embedded_min_dynamic_size() for t in self.subtyps])

    def is_complex_type(self):
        return True

    def selector_name(self):
        return "(" + ",".join(t.selector_name() for t in self.subtyps) + ")"



def test_string_map_keys(get_contract):
    code = """
f:HashMap[String[1], bool]
@external
def test() -> bool:
    a:String[1] = "a"
    b:String[1] = "b"
    self.f[a] = True
    return self.f[b]  # should return False
    """
    c = get_contract(code)
    c.test()
    assert c.test() is False


def test_string_map_keys_literals(get_contract):
    code = """
f:HashMap[String[1], bool]
@external
def test() -> bool:
    self.f["a"] = True
    return self.f["b"]  # should return False
    """
    c = get_contract(code)
    assert c.test() is False



import pytest

from vyper import compiler
from vyper.exceptions import FunctionDeclarationException, StructureException

fail_list = [
    (
        """
@external
def foo() -> int128:
    pass  # missing return
    """,
        FunctionDeclarationException,
    ),
    (
        """
@external
def foo() -> int128:
    if False:
        return 123
    # missing return
    """,
        FunctionDeclarationException,
    ),
    (
        """
@external
def test() -> int128:
    if 1 == 1 :
        return 1
        if True:  # unreachable
            return 0
    else:
        assert msg.sender != msg.sender
    """,
        StructureException,
    ),
    (
        """
@internal
def valid_address(sender: address) -> bool:
    selfdestruct(sender)
    return True  # unreachable
    """,
        StructureException,
    ),
    (
        """
@internal
def valid_address(sender: address) -> bool:
    selfdestruct(sender)
    a: address = sender  # unreachable
    """,
        StructureException,
    ),
    (
        """
@internal
def valid_address(sender: address) -> bool:
    if sender == empty(address):
        selfdestruct(sender)
        _sender: address = sender  # unreachable
    else:
        return False
    """,
        StructureException,
    ),
    (
        """
@internal
def foo() -> bool:
    raw_revert(b"vyper")
    return True  # unreachable
    """,
        StructureException,
    ),
    (
        """
@internal
def foo() -> bool:
    raw_revert(b"vyper")
    x: uint256 = 3  # unreachable
    """,
        StructureException,
    ),
    (
        """
@internal
def foo(x: uint256) -> bool:
    if x == 2:
        raw_revert(b"vyper")
        a: uint256 = 3  # unreachable
    else:
        return False
    """,
        StructureException,
    ),
    (
        """
@internal
def foo():
    return
    return  # unreachable
    """,
        StructureException,
    ),
    (
        """
@internal
def foo() -> uint256:
    if block.number % 2 == 0:
        return 5
    elif block.number % 3 == 0:
        return 6
    else:
        return 10
    return 0  # unreachable
    """,
        StructureException,
    ),
    (
        """
@internal
def foo() -> uint256:
    for i: uint256 in range(10):
        if i == 11:
            return 1
        """,
        FunctionDeclarationException,
    ),
    (
        """
@internal
def foo() -> uint256:
    for i: uint256 in range(9):
        if i == 11:
            return 1
    if block.number % 2 == 0:
        return 1
        """,
        FunctionDeclarationException,
    ),
    (
        """
@internal
def foo() -> uint256:
    for i: uint256 in range(10):
        return 1
        pass  # unreachable
    return 5
        """,
        StructureException,
    ),
]


@pytest.mark.parametrize("bad_code,exc", fail_list)
def test_return_mismatch(bad_code, exc):
    with pytest.raises(exc):
        compiler.compile_code(bad_code)


valid_list = [
    """
@external
def foo() -> int128:
    return 123
    """,
    """
@external
def foo() -> int128:
    if True:
        return 123
    else:
        raise "test"
    """,
    """
@external
def foo() -> int128:
    if False:
        return 123
    else:
        selfdestruct(msg.sender)
    """,
    """
@external
def foo() -> int128:
    if False:
        return 123
    return 333
    """,
    """
@external
def test() -> int128:
    if 1 == 1 :
        return 1
    else:
        assert msg.sender != msg.sender
        return 0
    """,
    """
@external
def test() -> int128:
    x: bytes32 = empty(bytes32)
    if False:
        if False:
            return 0
        else:
            x = keccak256(x)
            return 1
    else:
        x = keccak256(x)
        return 1
    """,
    """
@external
def foo() -> int128:
    if True:
        return 123
    else:
        raw_revert(b"vyper")
    """,
    """
@external
def foo() -> int128:
    for i: uint256 in range(1):
        return 1
    return 0
    """,
]


@pytest.mark.parametrize("good_code", valid_list)
def test_return_success(good_code):
    assert compiler.compile_code(good_code) is not None



import pytest

INITIAL_VALUE = 4


@pytest.fixture(scope="module")
def storage_contract(get_contract):
    with open("examples/storage/storage.vy") as f:
        contract_code = f.read()
        # Pass constructor variables directly to the contract
        contract = get_contract(contract_code, INITIAL_VALUE)
    return contract


def test_initial_state(storage_contract):
    # Check if the constructor of the contract is set up properly
    assert storage_contract.storedData() == INITIAL_VALUE


def test_set(storage_contract):
    storage_contract.set(10)
    assert storage_contract.storedData() == 10  # Directly access storedData

    storage_contract.set(-5)
    assert storage_contract.storedData() == -5



from vyper.codegen.ir_node import IRnode
from vyper.venom.ir_node_to_venom import ir_node_to_venom


def test_simple():
    ir = IRnode.from_list(["calldatacopy", 32, 0, ["calldatasize"]])
    ir_node = IRnode.from_list(ir)
    venom = ir_node_to_venom(ir_node)
    assert venom is not None

    fn = list(venom.functions.values())[0]

    bb = fn.entry
    assert bb.instructions[0].opcode == "calldatasize"
    assert bb.instructions[1].opcode == "calldatacopy"


def test_simple_2():
    ir = [
        "seq",
        [
            "seq",
            [
                "mstore",
                ["add", 64, 0],
                [
                    "with",
                    "x",
                    ["calldataload", ["add", 4, 0]],
                    [
                        "with",
                        "ans",
                        ["add", "x", 1],
                        ["seq", ["assert", ["ge", "ans", "x"]], "ans"],
                    ],
                ],
            ],
        ],
        32,
    ]
    ir_node = IRnode.from_list(ir)
    venom = ir_node_to_venom(ir_node)
    assert venom is not None



import pytest

from vyper.compiler import compile_code
from vyper.exceptions import MemoryAllocationException


def test_memory_overflow():
    code = """
@external
def zzz(x: DynArray[uint256, 2**59]):  # 2**64 / 32 bytes per word == 2**59
    y: uint256[7] = [0,0,0,0,0,0,0]

    y[6] = y[5]
    """
    with pytest.raises(MemoryAllocationException):
        compile_code(code)



from vyper.codegen.ir_node import IRnode
from vyper.evm.opcodes import version_check
from vyper.ir import compile_ir, optimizer


def test_sha3_32():
    ir = ["sha3_32", 0]
    evm = ["PUSH1", 0, "PUSH1", 0, "MSTORE", "PUSH1", 32, "PUSH1", 0, "SHA3"]
    if version_check(begin="shanghai"):
        evm = ["PUSH0", "PUSH0", "MSTORE", "PUSH1", 32, "PUSH0", "SHA3"]
    assert compile_ir.compile_to_assembly(IRnode.from_list(ir)) == evm
    assert compile_ir.compile_to_assembly(optimizer.optimize(IRnode.from_list(ir))) == evm



from pathlib import PurePath

import pytest

from vyper.cli.vyper_json import get_inputs
from vyper.exceptions import JSONError
from vyper.utils import keccak256

FOO_CODE = """
import contracts.bar as Bar

@external
def foo(a: address) -> bool:
    return Bar(a).bar(1)
"""

BAR_CODE = """
@external
def bar(a: uint256) -> bool:
    return True
"""


def test_no_sources():
    with pytest.raises(KeyError):
        get_inputs({})


def test_contracts_urls():
    with pytest.raises(JSONError):
        get_inputs({"sources": {"foo.vy": {"urls": ["https://foo.code.com/"]}}})


def test_contracts_no_content_key():
    with pytest.raises(JSONError):
        get_inputs({"sources": {"foo.vy": FOO_CODE}})


def test_contracts_keccak():
    hash_ = keccak256(FOO_CODE.encode()).hex()

    input_json = {"sources": {"foo.vy": {"content": FOO_CODE, "keccak256": hash_}}}
    get_inputs(input_json)

    input_json["sources"]["foo.vy"]["keccak256"] = "0x" + hash_
    get_inputs(input_json)

    input_json["sources"]["foo.vy"]["keccak256"] = "0x1234567890"
    with pytest.raises(JSONError):
        get_inputs(input_json)


def test_contracts_outside_pwd():
    input_json = {"sources": {"../foo.vy": {"content": FOO_CODE}}}
    get_inputs(input_json)


def test_contract_collision():
    # ./foo.vy and foo.vy will resolve to the same path
    input_json = {"sources": {"./foo.vy": {"content": FOO_CODE}, "foo.vy": {"content": FOO_CODE}}}
    with pytest.raises(JSONError):
        get_inputs(input_json)


def test_contracts_return_value():
    input_json = {
        "sources": {"foo.vy": {"content": FOO_CODE}, "contracts/bar.vy": {"content": BAR_CODE}}
    }
    result = get_inputs(input_json)
    assert result == {
        PurePath("foo.vy"): {"content": FOO_CODE},
        PurePath("contracts/bar.vy"): {"content": BAR_CODE},
    }


BAR_ABI = [
    {
        "name": "bar",
        "outputs": [{"type": "bool", "name": "out"}],
        "inputs": [{"type": "uint256", "name": "a"}],
        "stateMutability": "nonpayable",
        "type": "function",
    }
]


# tests to get interfaces from input dicts


def test_interface_collision():
    input_json = {
        "sources": {"foo.vy": {"content": FOO_CODE}},
        "interfaces": {"bar.json": {"abi": BAR_ABI}, "bar.vy": {"content": BAR_CODE}},
    }
    with pytest.raises(JSONError):
        get_inputs(input_json)


def test_json_no_abi():
    input_json = {
        "sources": {"foo.vy": {"content": FOO_CODE}},
        "interfaces": {"bar.json": {"content": BAR_ABI}},
    }
    with pytest.raises(JSONError):
        get_inputs(input_json)


def test_vy_no_content():
    input_json = {
        "sources": {"foo.vy": {"content": FOO_CODE}},
        "interfaces": {"bar.vy": {"abi": BAR_CODE}},
    }
    with pytest.raises(JSONError):
        get_inputs(input_json)


def test_interfaces_output():
    input_json = {
        "sources": {"foo.vy": {"content": FOO_CODE}},
        "interfaces": {
            "bar.json": {"abi": BAR_ABI},
            "interface.folder/bar2.vy": {"content": BAR_CODE},
        },
    }
    result = get_inputs(input_json)
    assert result == {
        PurePath("foo.vy"): {"content": FOO_CODE},
        PurePath("bar.json"): {"abi": BAR_ABI},
        PurePath("interface.folder/bar2.vy"): {"content": BAR_CODE},
    }


# EIP-2678 -- not currently supported
@pytest.mark.xfail
def test_manifest_output():
    input_json = {"interfaces": {"bar.json": {"contractTypes": {"Bar": {"abi": BAR_ABI}}}}}
    result = get_inputs(input_json)
    assert isinstance(result, dict)
    assert result == {"Bar": {"type": "json", "code": BAR_ABI}}



from dataclasses import dataclass
from functools import cached_property
from typing import Optional

from vyper.codegen.context import Constancy, Context
from vyper.codegen.ir_node import IRnode
from vyper.codegen.memory_allocator import MemoryAllocator
from vyper.evm.opcodes import version_check
from vyper.semantics.types import VyperType
from vyper.semantics.types.function import ContractFunctionT, StateMutability
from vyper.semantics.types.module import ModuleT
from vyper.utils import MemoryPositions


@dataclass
class FrameInfo:
    frame_start: int
    frame_size: int
    frame_vars: dict[str, tuple[int, VyperType]]

    @property
    def mem_used(self):
        return self.frame_size + MemoryPositions.RESERVED_MEMORY


@dataclass
class _FuncIRInfo:
    func_t: ContractFunctionT
    gas_estimate: Optional[int] = None
    frame_info: Optional[FrameInfo] = None

    @property
    def visibility(self):
        return "internal" if self.func_t.is_internal else "external"

    @property
    def exit_sequence_label(self) -> str:
        return self.ir_identifier + "_cleanup"

    @cached_property
    def ir_identifier(self) -> str:
        argz = ",".join([str(argtyp) for argtyp in self.func_t.argument_types])

        name = self.func_t.name
        function_id = self.func_t._function_id
        assert function_id is not None

        # include module id in the ir identifier to disambiguate functions
        # with the same name but which come from different modules
        return f"{self.visibility} {function_id} {name}({argz})"

    def set_frame_info(self, frame_info: FrameInfo) -> None:
        # XXX: when can this happen?
        if self.frame_info is not None:
            assert frame_info == self.frame_info
        else:
            self.frame_info = frame_info

    @property
    # common entry point for external function with kwargs
    def external_function_base_entry_label(self) -> str:
        assert not self.func_t.is_internal, "uh oh, should be external"
        return self.ir_identifier + "_common"

    def internal_function_label(self, is_ctor_context: bool = False) -> str:
        f = self.func_t
        assert f.is_internal or f.is_constructor, "uh oh, should be internal"

        if f.is_constructor:
            # sanity check - imported init functions only callable from main init
            assert is_ctor_context

        suffix = "_deploy" if is_ctor_context else "_runtime"
        return self.ir_identifier + suffix


@dataclass
class EntryPointInfo:
    func_t: ContractFunctionT
    min_calldatasize: int  # the min calldata required for this entry point
    ir_node: IRnode  # the ir for this entry point

    def __post_init__(self):
        # sanity check ABI v2 properties guaranteed by the spec.
        # https://docs.soliditylang.org/en/v0.8.21/abi-spec.html#formal-specification-of-the-encoding states:  # noqa: E501
        # > Note that for any X, len(enc(X)) is a multiple of 32.
        assert self.min_calldatasize >= 4
        assert (self.min_calldatasize - 4) % 32 == 0


@dataclass
class ExternalFuncIR:
    entry_points: dict[str, EntryPointInfo]  # map from abi sigs to entry points
    common_ir: IRnode  # the "common" code for the function


@dataclass
class InternalFuncIR:
    func_ir: IRnode  # the code for the function


def init_ir_info(func_t: ContractFunctionT):
    # initialize IRInfo on the function
    func_t._ir_info = _FuncIRInfo(func_t)


def initialize_context(
    func_t: ContractFunctionT, module_ctx: ModuleT, is_ctor_context: bool = False
):
    init_ir_info(func_t)

    # calculate starting frame
    callees = func_t.called_functions
    # we start our function frame from the largest callee frame
    max_callee_frame_size = 0
    for c_func_t in callees:
        frame_info = c_func_t._ir_info.frame_info
        max_callee_frame_size = max(max_callee_frame_size, frame_info.frame_size)

    allocate_start = max_callee_frame_size + MemoryPositions.RESERVED_MEMORY

    memory_allocator = MemoryAllocator(allocate_start)

    return Context(
        vars_=None,
        module_ctx=module_ctx,
        memory_allocator=memory_allocator,
        constancy=Constancy.Mutable if func_t.is_mutable else Constancy.Constant,
        func_t=func_t,
        is_ctor_context=is_ctor_context,
    )


def tag_frame_info(func_t, context):
    frame_size = context.memory_allocator.size_of_mem - MemoryPositions.RESERVED_MEMORY
    frame_start = context.starting_memory

    frame_info = FrameInfo(frame_start, frame_size, context.vars)
    func_t._ir_info.set_frame_info(frame_info)

    return frame_info


def get_nonreentrant_lock(func_t):
    if not func_t.nonreentrant:
        return ["pass"], ["pass"]

    nkey = func_t.reentrancy_key_position.position

    LOAD, STORE = "sload", "sstore"
    if version_check(begin="cancun"):
        LOAD, STORE = "tload", "tstore"
        # for tload/tstore we don't need to care about net gas metering,
        # choose small constants (e.g. 0 can be replaced by PUSH0)
        final_value, temp_value = 0, 1
    else:
        # any nonzero values can work here (see pricing as of net gas
        # metering); these values are chosen so that downgrading to the
        # 0,1 scheme (if it is somehow necessary) is safe.
        final_value, temp_value = 3, 2

    check_notset = ["assert", ["ne", temp_value, [LOAD, nkey]]]

    if func_t.mutability == StateMutability.VIEW:
        return [check_notset], [["seq"]]

    else:
        pre = ["seq", check_notset, [STORE, nkey, temp_value]]
        post = [STORE, nkey, final_value]
        return [pre], [post]



import pytest


@pytest.fixture(autouse=True)
def set_balance(env):
    for a in env.accounts[:7]:
        env.set_balance(a, 10**10)


# TODO: check, this is probably redundant with examples/test_crowdfund.py
def test_crowdfund(env, get_contract):
    crowdfund = """

struct Funder:
    sender: address
    value: uint256
funders: HashMap[int128, Funder]
nextFunderIndex: int128
beneficiary: address
deadline: public(uint256)
goal: public(uint256)
refundIndex: int128
timelimit: public(uint256)

@deploy
def __init__(_beneficiary: address, _goal: uint256, _timelimit: uint256):
    self.beneficiary = _beneficiary
    self.deadline = block.timestamp + _timelimit
    self.timelimit = _timelimit
    self.goal = _goal

@external
@payable
def participate():
    assert block.timestamp < self.deadline
    nfi: int128 = self.nextFunderIndex
    self.funders[nfi].sender = msg.sender
    self.funders[nfi].value = msg.value
    self.nextFunderIndex = nfi + 1

@external
@view
def expired() -> bool:
    return block.timestamp >= self.deadline

@external
@view
def block_timestamp() -> uint256:
    return block.timestamp

@external
@view
def reached() -> bool:
    return self.balance >= self.goal

@external
def finalize():
    assert block.timestamp >= self.deadline and self.balance >= self.goal
    selfdestruct(self.beneficiary)

@external
def refund():
    ind: int128 = self.refundIndex
    for i: int128 in range(ind, ind + 30, bound=30):
        if i >= self.nextFunderIndex:
            self.refundIndex = self.nextFunderIndex
            return
        send(self.funders[i].sender, self.funders[i].value)
        self.funders[i].sender = 0x0000000000000000000000000000000000000000
        self.funders[i].value = 0
    self.refundIndex = ind + 30

    """
    a0, a1, a2, a3, a4, a5, a6 = env.accounts[:7]

    c = get_contract(crowdfund, *[a1, 50, 60])
    start_timestamp = env.timestamp

    c.participate(value=5)
    assert c.timelimit() == 60
    assert c.deadline() - start_timestamp == 60
    assert not c.expired()
    assert not c.reached()
    c.participate(value=49)
    assert c.reached()
    pre_bal = env.get_balance(a1)
    env.timestamp += 100
    assert c.expired()
    c.finalize()
    post_bal = env.get_balance(a1)
    assert post_bal - pre_bal == 54

    c = get_contract(crowdfund, *[a1, 50, 60])
    c.participate(value=1, sender=a3)
    c.participate(value=2, sender=a4)
    c.participate(value=3, sender=a5)
    c.participate(value=4, sender=a6)
    env.timestamp += 100
    assert c.expired()
    assert not c.reached()
    pre_bals = [env.get_balance(x) for x in [a3, a4, a5, a6]]
    c.refund()
    post_bals = [env.get_balance(x) for x in [a3, a4, a5, a6]]
    assert [y - x for x, y in zip(pre_bals, post_bals)] == [1, 2, 3, 4]


def test_crowdfund2(env, get_contract):
    crowdfund2 = """
struct Funder:
    sender: address
    value: uint256

funders: HashMap[int128, Funder]
nextFunderIndex: int128
beneficiary: address
deadline: public(uint256)
goal: uint256
refundIndex: int128
timelimit: public(uint256)

@deploy
def __init__(_beneficiary: address, _goal: uint256, _timelimit: uint256):
    self.beneficiary = _beneficiary
    self.deadline = block.timestamp + _timelimit
    self.timelimit = _timelimit
    self.goal = _goal

@external
@payable
def participate():
    assert block.timestamp < self.deadline
    nfi: int128 = self.nextFunderIndex
    self.funders[nfi] = Funder(sender=msg.sender, value=msg.value)
    self.nextFunderIndex = nfi + 1

@external
@view
def expired() -> bool:
    return block.timestamp >= self.deadline

@external
@view
def block_timestamp() -> uint256:
    return block.timestamp

@external
@view
def reached() -> bool:
    return self.balance >= self.goal

@external
def finalize():
    assert block.timestamp >= self.deadline and self.balance >= self.goal
    selfdestruct(self.beneficiary)

@external
def refund():
    ind: int128 = self.refundIndex
    for i: int128 in range(ind, ind + 30, bound=30):
        if i >= self.nextFunderIndex:
            self.refundIndex = self.nextFunderIndex
            return
        send(self.funders[i].sender, self.funders[i].value)
        self.funders[i] = empty(Funder)
    self.refundIndex = ind + 30

    """
    a0, a1, a2, a3, a4, a5, a6 = env.accounts[:7]
    c = get_contract(crowdfund2, *[a1, 50, 60])

    c.participate(value=5)
    env.timestamp += 1  # make sure auction has started
    assert c.timelimit() == 60
    assert c.deadline() - c.block_timestamp() == 59
    assert not c.expired()
    assert not c.reached()
    c.participate(value=49)
    assert c.reached()
    pre_bal = env.get_balance(a1)
    env.timestamp += 100
    assert c.expired()
    c.finalize()
    post_bal = env.get_balance(a1)
    assert post_bal - pre_bal == 54

    c = get_contract(crowdfund2, *[a1, 50, 60])
    c.participate(value=1, sender=a3)
    c.participate(value=2, sender=a4)
    c.participate(value=3, sender=a5)
    c.participate(value=4, sender=a6)
    env.timestamp += 100
    assert c.expired()
    assert not c.reached()
    pre_bals = [env.get_balance(x) for x in [a3, a4, a5, a6]]
    c.refund()
    post_bals = [env.get_balance(x) for x in [a3, a4, a5, a6]]
    assert [y - x for x, y in zip(pre_bals, post_bals)] == [1, 2, 3, 4]



import pytest

from vyper import compile_code
from vyper.exceptions import ArgumentException

fail_list = [
    """
@external
def foo():
    x: uint256 = as_wei_value(5, "vader")
    """,
    """
@external
def foo(x: int128, x: int128): pass
    """,
    """
@external
def foo(x): pass
    """,
    """
@external
def foo() -> int128:
    return as_wei_value(10)
    """,
    """
@external
def foo():
    x: bytes32 = keccak256("moose", 3)
    """,
    """
@external
def foo():
    x: Bytes[4] = raw_call(0x1234567890123456789012345678901234567890, outsize=4)
    """,
    """
@external
def foo():
    x: Bytes[4] = raw_call(
        0x1234567890123456789012345678901234567890, b"cow", gas=111111, outsize=4, moose=9
    )
    """,
    """
@external
def foo():
    x: Bytes[4] = create_minimal_proxy_to(0x1234567890123456789012345678901234567890, outsize=4)
    """,
    """
x: public()
    """,
    """
@external
def foo():
    raw_log([], b"cow", "dog")
    """,
    """
@external
def foo():
    x: Bytes[10] = concat(b"")
    """,
    """
@external
def foo():
    x: Bytes[4] = create_minimal_proxy_to(0x1234567890123456789012345678901234567890, b"cow")
    """,
    """
@external
def foo():
    a: uint256 = min()
    """,
    """
@external
def foo():
    a: uint256 = min(1)
    """,
    """
@external
def foo():
    a: uint256 = min(1, 2, 3)
    """,
    """
@external
def foo():
    for i: uint256 in range():
        pass
    """,
    """
@external
def foo():
    for i: uint256 in range(1, 2, 3, 4):
        pass
    """,
]


@pytest.mark.parametrize("bad_code", fail_list)
def test_function_declaration_exception(bad_code):
    with pytest.raises(ArgumentException):
        compile_code(bad_code)



from typing import Optional

from vyper.exceptions import CompilerPanic
from vyper.utils import OrderedSet
from vyper.venom.analysis.analysis import IRAnalysesCache
from vyper.venom.analysis.dominators import DominatorTreeAnalysis
from vyper.venom.basicblock import IRBasicBlock, IRInstruction, IRLabel, IRLiteral, IRVariable
from vyper.venom.context import IRContext
from vyper.venom.function import IRFunction
from vyper.venom.passes.make_ssa import MakeSSA


def _add_bb(
    fn: IRFunction, label: IRLabel, cfg_outs: list[IRLabel], bb: Optional[IRBasicBlock] = None
) -> IRBasicBlock:
    bb = bb if bb is not None else IRBasicBlock(label, fn)
    fn.append_basic_block(bb)
    cfg_outs_len = len(cfg_outs)
    if cfg_outs_len == 0:
        bb.append_instruction("stop")
    elif cfg_outs_len == 1:
        bb.append_instruction("jmp", cfg_outs[0])
    elif cfg_outs_len == 2:
        bb.append_instruction("jnz", IRLiteral(1), cfg_outs[0], cfg_outs[1])
    else:
        raise CompilerPanic("Invalid number of CFG outs")
    return bb


def _make_test_ctx():
    lab = [IRLabel(str(i)) for i in range(0, 9)]

    ctx = IRContext()
    fn = ctx.create_function(lab[1].value)

    fn.entry.append_instruction("jmp", lab[2])

    _add_bb(fn, lab[7], [])
    _add_bb(fn, lab[6], [lab[7], lab[2]])
    _add_bb(fn, lab[5], [lab[6], lab[3]])
    _add_bb(fn, lab[4], [lab[6]])
    _add_bb(fn, lab[3], [lab[5]])
    _add_bb(fn, lab[2], [lab[3], lab[4]])

    return fn


def test_deminator_frontier_calculation():
    fn = _make_test_ctx()
    bb1, bb2, bb3, bb4, bb5, bb6, bb7 = [fn.get_basic_block(str(i)) for i in range(1, 8)]

    ac = IRAnalysesCache(fn)
    dom = ac.request_analysis(DominatorTreeAnalysis)
    df = dom.dominator_frontiers

    assert len(df[bb1]) == 0, df[bb1]
    assert df[bb2] == OrderedSet({bb2}), df[bb2]
    assert df[bb3] == OrderedSet({bb3, bb6}), df[bb3]
    assert df[bb4] == OrderedSet({bb6}), df[bb4]
    assert df[bb5] == OrderedSet({bb3, bb6}), df[bb5]
    assert df[bb6] == OrderedSet({bb2}), df[bb6]
    assert len(df[bb7]) == 0, df[bb7]


def test_phi_placement():
    fn = _make_test_ctx()
    bb1, bb2, bb3, bb4, bb5, bb6, bb7 = [fn.get_basic_block(str(i)) for i in range(1, 8)]

    x = IRVariable("%x")
    bb1.insert_instruction(IRInstruction("mload", [IRLiteral(0)], x), 0)
    bb2.insert_instruction(IRInstruction("add", [x, IRLiteral(1)], x), 0)
    bb7.insert_instruction(IRInstruction("mstore", [x, IRLiteral(0)]), 0)

    ac = IRAnalysesCache(fn)
    MakeSSA(ac, fn).run_pass()



from contextlib import contextmanager
from random import Random
from typing import Generator

import hypothesis
import pytest
from eth_keys.datatypes import PrivateKey
from hexbytes import HexBytes

import vyper.evm.opcodes as evm_opcodes
from tests.evm_backends.base_env import BaseEnv, ExecutionReverted
from tests.evm_backends.pyevm_env import PyEvmEnv
from tests.evm_backends.revm_env import RevmEnv
from tests.utils import working_directory
from vyper import compiler
from vyper.codegen.ir_node import IRnode
from vyper.compiler.input_bundle import FilesystemInputBundle, InputBundle
from vyper.compiler.settings import OptimizationLevel, Settings, set_global_settings
from vyper.exceptions import EvmVersionException
from vyper.ir import compile_ir, optimizer
from vyper.utils import keccak256

############
# PATCHING #
############


# disable hypothesis deadline globally
hypothesis.settings.register_profile("ci", deadline=None)
hypothesis.settings.load_profile("ci")


def pytest_addoption(parser):
    parser.addoption(
        "--optimize",
        choices=["codesize", "gas", "none"],
        default="gas",
        help="change optimization mode",
    )
    parser.addoption("--enable-compiler-debug-mode", action="store_true")
    parser.addoption("--experimental-codegen", action="store_true")
    parser.addoption("--tracing", action="store_true")

    parser.addoption(
        "--evm-version",
        choices=list(evm_opcodes.EVM_VERSIONS.keys()),
        default="shanghai",
        help="set evm version",
    )

    parser.addoption(
        "--evm-backend", choices=["py-evm", "revm"], default="revm", help="set evm backend"
    )


@pytest.fixture(scope="module")
def output_formats():
    output_formats = compiler.OUTPUT_FORMATS.copy()

    to_drop = ("bb", "bb_runtime", "cfg", "cfg_runtime", "archive", "archive_b64", "solc_json")
    for s in to_drop:
        del output_formats[s]

    return output_formats


@pytest.fixture(scope="session")
def optimize(pytestconfig):
    flag = pytestconfig.getoption("optimize")
    return OptimizationLevel.from_string(flag)


@pytest.fixture(scope="session")
def debug(pytestconfig):
    debug = pytestconfig.getoption("enable_compiler_debug_mode")
    assert isinstance(debug, bool)
    return debug


@pytest.fixture(scope="session")
def experimental_codegen(pytestconfig):
    ret = pytestconfig.getoption("experimental_codegen")
    assert isinstance(ret, bool)
    return ret


@pytest.fixture(autouse=True)
def check_venom_xfail(request, experimental_codegen):
    if not experimental_codegen:
        return

    marker = request.node.get_closest_marker("venom_xfail")
    if marker is None:
        return

    # https://github.com/okken/pytest-runtime-xfail?tab=readme-ov-file#alternatives
    request.node.add_marker(pytest.mark.xfail(strict=True, **marker.kwargs))


@pytest.fixture
def venom_xfail(request, experimental_codegen):
    def _xfail(*args, **kwargs):
        if not experimental_codegen:
            return
        request.node.add_marker(pytest.mark.xfail(*args, strict=True, **kwargs))

    return _xfail


@pytest.fixture(scope="session")
def evm_version(pytestconfig):
    # note: configure the evm version that we emit code for.
    # The env will read this fixture and apply the evm version there.
    return pytestconfig.getoption("evm_version")


@pytest.fixture(scope="session")
def evm_backend(pytestconfig):
    backend_str = pytestconfig.getoption("evm_backend")
    return {"py-evm": PyEvmEnv, "revm": RevmEnv}[backend_str]


@pytest.fixture(scope="session")
def tracing(pytestconfig):
    return pytestconfig.getoption("tracing")


@pytest.fixture
def chdir_tmp_path(tmp_path):
    # this is useful for when you want imports to have relpaths
    with working_directory(tmp_path):
        yield


# CMC 2024-03-01 this doesn't need to be a fixture
@pytest.fixture
def keccak():
    return keccak256


@pytest.fixture
def make_file(tmp_path):
    # writes file_contents to file_name, creating it in the
    # tmp_path directory. returns final path.
    def fn(file_name, file_contents):
        path = tmp_path / file_name
        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open("w") as f:
            f.write(file_contents)

        return path

    return fn


# this can either be used for its side effects (to prepare a call
# to get_contract), or the result can be provided directly to
# compile_code / CompilerData.
@pytest.fixture
def make_input_bundle(tmp_path, make_file):
    def fn(sources_dict):
        for file_name, file_contents in sources_dict.items():
            make_file(file_name, file_contents)
        return FilesystemInputBundle([tmp_path])

    return fn


# for tests which just need an input bundle, doesn't matter what it is
@pytest.fixture
def dummy_input_bundle():
    return InputBundle([])


@pytest.fixture(scope="module")
def gas_limit():
    # set absurdly high gas limit so that london basefee never adjusts
    # (note: 2**63 - 1 is max that py-evm allows)
    return 10**10


@pytest.fixture(scope="module")
def account_keys():
    random = Random(b"vyper")
    return [PrivateKey(random.randbytes(32)) for _ in range(10)]


@pytest.fixture(scope="module")
def env(gas_limit, evm_version, evm_backend, tracing, account_keys) -> BaseEnv:
    return evm_backend(
        gas_limit=gas_limit,
        tracing=tracing,
        block_number=1,
        evm_version=evm_version,
        account_keys=account_keys,
    )


@pytest.fixture
def get_contract_from_ir(env, optimize):
    def ir_compiler(ir, *args, **kwargs):
        ir = IRnode.from_list(ir)
        if kwargs.pop("optimize", optimize) != OptimizationLevel.NONE:
            ir = optimizer.optimize(ir)

        assembly = compile_ir.compile_to_assembly(ir, optimize=optimize)
        bytecode, _ = compile_ir.assembly_to_evm(assembly)

        abi = kwargs.pop("abi", [])
        return env.deploy(abi, bytecode, *args, **kwargs)

    return ir_compiler


@pytest.fixture(scope="module", autouse=True)
def compiler_settings(optimize, experimental_codegen, evm_version, debug):
    compiler.settings.DEFAULT_ENABLE_DECIMALS = True
    settings = Settings(
        optimize=optimize,
        evm_version=evm_version,
        experimental_codegen=experimental_codegen,
        debug=debug,
    )
    set_global_settings(settings)
    return settings


@pytest.fixture(scope="module")
def get_contract(env, optimize, output_formats, compiler_settings):
    def fn(source_code, *args, **kwargs):
        if "override_opt_level" in kwargs:
            kwargs["compiler_settings"] = Settings(
                **dict(compiler_settings.__dict__, optimize=kwargs.pop("override_opt_level"))
            )
        return env.deploy_source(source_code, output_formats, *args, **kwargs)

    return fn


@pytest.fixture(scope="module")
def deploy_blueprint_for(env, output_formats):
    def fn(source_code, *args, **kwargs):
        # we don't pass any settings, but it will pick up the global settings
        return env.deploy_blueprint(source_code, output_formats, *args, **kwargs)

    return fn


@pytest.fixture(scope="module")
def get_logs(env):
    return env.get_logs


# TODO: this should not be a fixture.
# remove me and replace all uses with `with pytest.raises`.
@pytest.fixture
def assert_compile_failed(tx_failed):
    def assert_compile_failed(function_to_test, exception=Exception):
        with tx_failed(exception):
            function_to_test()

    return assert_compile_failed


@pytest.fixture
def create2_address_of(keccak):
    def _f(_addr, _salt, _initcode):
        prefix = HexBytes("0xff")
        addr = HexBytes(_addr)
        salt = HexBytes(_salt)
        initcode = HexBytes(_initcode)
        return keccak(prefix + addr + salt + keccak(initcode))[12:]

    return _f


@pytest.fixture
def side_effects_contract(get_contract):
    def generate(ret_type):
        """
        Generates a Vyper contract with an external `foo()` function, which
        returns the specified return value of the specified return type, for
        testing side effects using the `assert_side_effects_invoked` fixture.
        """
        code = f"""
counter: public(uint256)

@external
def foo(s: {ret_type}) -> {ret_type}:
    self.counter += 1
    return s
    """
        return get_contract(code)

    return generate


@pytest.fixture
def assert_side_effects_invoked():
    def assert_side_effects_invoked(side_effects_contract, side_effects_trigger, n=1):
        start_value = side_effects_contract.counter()

        side_effects_trigger()

        end_value = side_effects_contract.counter()
        assert end_value == start_value + n

    return assert_side_effects_invoked


# should probably be renamed since there is no longer a transaction object
@pytest.fixture(scope="module")
def tx_failed(env):
    @contextmanager
    def fn(exception=ExecutionReverted, exc_text=None):
        with pytest.raises(exception) as excinfo:
            yield

        if exc_text:
            # TODO test equality
            assert exc_text in str(excinfo.value), (exc_text, excinfo.value)

    return fn


@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_call(item) -> Generator:
    marker = item.get_closest_marker("requires_evm_version")
    if marker:
        assert len(marker.args) == 1
        version = marker.args[0]
        if not evm_opcodes.version_check(begin=version):
            item.add_marker(
                pytest.mark.xfail(reason="Wrong EVM version", raises=EvmVersionException)
            )

    # Isolate tests by reverting the state of the environment after each test
    env = item.funcargs.get("env")
    if env:
        with env.anchor():
            yield
    else:
        yield



import pytest

from vyper.exceptions import InvalidType, TypeMismatch

fail_list = [
    (
        """
@external
def foo() -> uint256:
    return extract32(b"cowcowcowcowcowccowcowcowcowcowccowcowcowcowcowccowcowcowcowcowc", 0)
    """,
        TypeMismatch,  # default return type is bytes32
    ),
    (
        """
@external
def foo(inp: address) -> int128:
    return extract32(inp, 0, output_type=int128)
    """,
        TypeMismatch,  # can't extract32 on an address
    ),
    (
        """
@external
def foo(inp: Bytes[32]) -> int128:
    b: int128 = 1
    return extract32(inp, b, output_type=int128)
    """,
        TypeMismatch,  # `start` must be an unsigned integer
    ),
    (
        """
@external
def foo(inp: Bytes[32]) -> int128:
    return extract32(inp, -1, output_type=int128)
    """,
        TypeMismatch,  # `start` cannot be negative
    ),
    (
        """
@external
def foo(inp: Bytes[32]) -> bool:
    return extract32(inp, 0, output_type=bool)
    """,
        InvalidType,  # output_type can't be bool
    ),
]


@pytest.mark.parametrize("bad_code,exc", fail_list)
def test_extract32_fail(assert_compile_failed, get_contract, bad_code, exc):
    assert_compile_failed(lambda: get_contract(bad_code), exc)


valid_list = [
    """
@external
def foo() -> uint256:
    return extract32(
        b"cowcowcowcowcowccowcowcowcowcowccowcowcowcowcowccowcowcowcowcowc",
        0,
        output_type=uint256
    )
    """,
    """
x: Bytes[100]
@external
def foo() -> uint256:
    self.x = b"cowcowcowcowcowccowcowcowcowcowccowcowcowcowcowccowcowcowcowcowc"
    return extract32(self.x, 0, output_type=uint256)
    """,
    """
x: Bytes[100]
@external
def foo() -> uint256:
    self.x = b"cowcowcowcowcowccowcowcowcowcowccowcowcowcowcowccowcowcowcowcowc"
    return extract32(self.x, 1, output_type=uint256)
""",
]


@pytest.mark.parametrize("good_code", valid_list)
def test_extract32_success(get_contract, good_code):
    assert get_contract(good_code) is not None



import pytest

from vyper.compiler import compile_code
from vyper.exceptions import CallViolation


@pytest.mark.parametrize(
    "source",
    [
        """
interface PiggyBank:
    def deposit(): nonpayable

piggy: PiggyBank

@external
def foo():
    extcall self.piggy.deposit()
    """,
        # You don't have to send value in a payable call
        """
interface PiggyBank:
    def deposit(): payable

piggy: PiggyBank

@external
def foo():
    extcall self.piggy.deposit()
    """,
    ],
)
def test_payable_call_compiles(source, get_contract):
    _ = compile_code(source)


@pytest.mark.parametrize(
    "source",
    [
        """
interface PiggyBank:
    def deposit(): nonpayable

piggy: PiggyBank

@external
def foo():
    # sends value to nonpayable function
    extcall self.piggy.deposit(value=self.balance)
    """
    ],
)
def test_payable_compile_fail(source, get_contract, assert_compile_failed):
    with pytest.raises(CallViolation):
        compile_code(source)


nonpayable_code = [
    """
# single function, nonpayable
@external
def foo() -> bool:
    return True
    """,
    """
# multiple functions, one is payable
@external
def foo() -> bool:
    return True

@payable
@external
def bar() -> bool:
    return True
    """,
    """
# multiple functions, nonpayable
@external
def foo() -> bool:
    return True

@external
def bar() -> bool:
    return True
    """,
    """
# multiple functions and default func, nonpayable
@external
def foo() -> bool:
    return True

@external
def bar() -> bool:
    return True

@external
def __default__():
    pass
    """,
    """
    # multiple functions and default func, payable
@external
def foo() -> bool:
    return True

@external
def bar() -> bool:
    return True

@external
@payable
def __default__():
    pass
    """,
    """
# multiple functions, nonpayable (view)
@external
def foo() -> bool:
    return True

@view
@external
def bar() -> bool:
    return True
    """,
    """
# payable init function
@deploy
@payable
def __init__():
    a: int128 = 1

@external
def foo() -> bool:
    return True
    """,
    """
# payable default function
@external
@payable
def __default__():
    a: int128 = 1

@external
def foo() -> bool:
    return True
    """,
    """
# payable default function and other function
@external
@payable
def __default__():
    a: int128 = 1

@external
def foo() -> bool:
    return True

@external
@payable
def bar() -> bool:
    return True
    """,
    """
# several functions, one payable
@external
def foo() -> bool:
    return True

@payable
@external
def bar() -> bool:
    return True

@external
def baz() -> bool:
    return True
    """,
]


@pytest.mark.parametrize("code", nonpayable_code)
def test_nonpayable_runtime_assertion(env, keccak, tx_failed, get_contract, code):
    c = get_contract(code)
    env.set_balance(env.deployer, 10**18)

    c.foo(value=0)
    sig = keccak("foo()".encode()).hex()[:10]
    with tx_failed():
        env.message_call(c.address, data=sig, value=10**18)


payable_code = [
    """
# single function, payable
@payable
@external
def foo() -> bool:
    return True
    """,
    """
# two functions, one is payable
@payable
@external
def foo() -> bool:
    return True

@external
def bar() -> bool:
    return True
    """,
    """
# two functions, payable
@payable
@external
def foo() -> bool:
    return True

@payable
@external
def bar() -> bool:
    return True
    """,
    """
# two functions, one nonpayable (view)
@payable
@external
def foo() -> bool:
    return True

@view
@external
def bar() -> bool:
    return True
    """,
    """
# several functions, all payable
@payable
@external
def foo() -> bool:
    return True

@payable
@external
def bar() -> bool:
    return True

@payable
@external
def baz() -> bool:
    return True
    """,
    """
# several functions, one payable
@payable
@external
def foo() -> bool:
    return True

@external
def bar() -> bool:
    return True

@external
def baz() -> bool:
    return True
    """,
    """
# several functions, two payable
@payable
@external
def foo() -> bool:
    return True

@external
def bar() -> bool:
    return True

@payable
@external
def baz() -> bool:
    return True
    """,
    """
# init function
@deploy
def __init__():
    a: int128 = 1

@payable
@external
def foo() -> bool:
    return True
    """,
    """
# default function
@external
def __default__():
    a: int128 = 1

@external
@payable
def foo() -> bool:
    return True
    """,
    """
# payable default function
@external
@payable
def __default__():
    a: int128 = 1

@external
@payable
def foo() -> bool:
    return True
    """,
    """
# payable default function and nonpayable other function
@external
@payable
def __default__():
    a: int128 = 1

@external
@payable
def foo() -> bool:
    return True

@external
def bar() -> bool:
    return True
    """,
]


@pytest.mark.parametrize("code", payable_code)
def test_payable_runtime_assertion(env, get_contract, code):
    c = get_contract(code)
    env.set_balance(env.deployer, 10**18)
    c.foo(value=10**18)
    c.foo(value=0)


def test_payable_default_func_invalid_calldata(get_contract, env):
    code = """
@external
def foo() -> bool:
    return True

@payable
@external
def __default__():
    pass
    """
    c = get_contract(code)
    env.set_balance(env.deployer, 100)
    env.message_call(c.address, value=100, data="0x12345678")


def test_nonpayable_default_func_invalid_calldata(get_contract, env, tx_failed):
    code = """
@external
@payable
def foo() -> bool:
    return True

@external
def __default__():
    pass
    """

    c = get_contract(code)
    env.message_call(c.address, value=0, data="0x12345678")
    env.set_balance(env.deployer, 100)
    with tx_failed():
        env.message_call(c.address, value=100, data="0x12345678")


def test_batch_nonpayable(get_contract, env, tx_failed):
    code = """
@external
def foo() -> bool:
    return True

@external
def __default__():
    pass
    """

    c = get_contract(code)
    env.message_call(c.address, value=0, data="0x12345678")
    data = bytes([1, 2, 3, 4])
    for i in range(5):
        calldata = "0x" + data[:i].hex()
        env.set_balance(env.deployer, 100)
        with tx_failed():
            env.message_call(c.address, value=100, data=calldata)



import pytest

from vyper import compiler
from vyper.exceptions import (
    FunctionDeclarationException,
    InvalidOperation,
    StructureException,
    TypeMismatch,
    UndeclaredDefinition,
    UnknownAttribute,
)

# These functions register test cases
# for pytest functions at the end
fail_list = []


def must_fail(code, exception):
    fail_list.append((code, exception))


pass_list = []


def must_succeed(code):
    pass_list.append(code)


# TEST CASES


must_succeed(
    """
x: int128[3]
"""
)

must_succeed(
    """
@external
def foo(x: int128): pass
"""
)

must_succeed(
    """
@external
def foo():
    x: int128 = 0
    x = 5
"""
)

must_succeed(
    """
@external
def foo():
    x: int128  = 5
"""
)

must_fail(
    """
@external
def foo():
    x: int128 = 5
    x = 0x1234567890123456789012345678901234567890
""",
    TypeMismatch,
)

must_fail(
    """
@external
def foo():
    x: int128 = 5
    x = 3.5
""",
    TypeMismatch,
)

must_succeed(
    """
@external
def foo():
    x: int128 = 5
    x = 3
"""
)

must_succeed(
    """
b: int128
@external
def foo():
    self.b = 7
"""
)

must_fail(
    """
b: int128
@external
def foo():
    self.b = 7.5
""",
    TypeMismatch,
)

must_succeed(
    """
b: decimal
@external
def foo():
    self.b = 7.5
"""
)

must_succeed(
    """
b: decimal
@external
def foo():
    self.b = 7.0
"""
)

must_fail(
    """
b: int128[5]
@external
def foo():
    self.b = 7
""",
    TypeMismatch,
)

must_succeed(
    """
b: HashMap[int128, int128]
@external
def foo():
    x: int128 = self.b[5]
"""
)

must_fail(
    """
b: HashMap[uint256, uint256]
@external
def foo():
    x: int128 = self.b[-5]
""",
    TypeMismatch,
)

must_fail(
    """
b: HashMap[int128, int128]
@external
def foo():
    x: int128 = self.b[5.7]
""",
    TypeMismatch,
)

must_succeed(
    """
b: HashMap[decimal, int128]
@external
def foo():
    x: int128 = self.b[5.0]
"""
)

must_fail(
    """
b: HashMap[int128, int128]
@external
def foo():
    self.b[3] = 5.6
""",
    TypeMismatch,
)

must_succeed(
    """
b: HashMap[int128, int128]
@external
def foo():
    self.b[3] = -5
"""
)

must_succeed(
    """
b: HashMap[int128, int128]
@external
def foo():
    self.b[-3] = 5
"""
)

must_succeed(
    """
@external
def foo():
    x: int128[5] = [0, 0, 0, 0, 0]
    z: int128 = x[2]
"""
)

must_succeed(
    """
x: int128
@external
def foo():
    self.x = 5
"""
)

must_succeed(
    """
x: int128
@internal
def foo():
    self.x = 5
"""
)

must_fail(
    """
bar: int128[3]
@external
def foo():
    self.bar = 5
""",
    TypeMismatch,
)

must_succeed(
    """
bar: int128[3]
@external
def foo():
    self.bar[0] = 5
"""
)

must_fail(
    """
@external
def foo() -> address:
    return [1, 2, 3]
""",
    TypeMismatch,
)

must_fail(
    """
@external
def baa() -> decimal:
    return 2.0**2
""",
    TypeMismatch,
)

must_succeed(
    """
@external
def foo():
    raise "fail"
"""
)

must_succeed(
    """
@internal
def foo():
    pass

@external
def goo():
    self.foo()
"""
)

must_succeed(
    """
@external
def foo():
    MOOSE: int128 = 45
"""
)

must_fail(
    """
@external
def foo():
    x: address = -self
""",
    InvalidOperation,
)

must_fail(
    """
@external
def foo() -> int128:
    return
""",
    FunctionDeclarationException,
)

must_fail(
    """
@external
def foo():
    return 3
""",
    FunctionDeclarationException,
)

must_fail(
    """
@external
def foo():
    suicide(msg.sender)
    """,
    UndeclaredDefinition,
)

must_succeed(
    '''
@external
def sum(a: int128, b: int128) -> int128:
    """
    Sum two signed integers.
    """
    return a + b
'''
)

must_fail(
    """
@external
def a():
    "Behold me mortal, for I am a DOCSTRING!"
    "Alas, I am but a mere string."
""",
    StructureException,
)

must_fail(
    """
struct StructX:
    x: int128

@external
def a():
    x: int128 = StructX(y=1)
""",
    UnknownAttribute,
)

must_fail(
    """
a: HashMap
""",
    StructureException,
)


@pytest.mark.parametrize("bad_code,exception_type", fail_list)
def test_compilation_fails_with_exception(bad_code, exception_type):
    with pytest.raises(exception_type):
        compiler.compile_code(bad_code)


@pytest.mark.parametrize("good_code", pass_list)
def test_compilation_succeeds(good_code):
    assert compiler.compile_code(good_code) is not None



#!/usr/bin/env python3

import argparse
import json
import sys
import warnings
from pathlib import Path, PurePath
from typing import Any, Callable, Hashable, Optional

import vyper
from vyper.compiler.input_bundle import FileInput, JSONInputBundle
from vyper.compiler.settings import OptimizationLevel, Settings
from vyper.evm.opcodes import EVM_VERSIONS
from vyper.exceptions import JSONError
from vyper.utils import OrderedSet, keccak256

TRANSLATE_MAP = {
    "abi": "abi",
    "ast": "ast_dict",
    "annotated_ast": "annotated_ast_dict",
    "devdoc": "devdoc",
    "evm.methodIdentifiers": "method_identifiers",
    "evm.bytecode.object": "bytecode",
    "evm.bytecode.opcodes": "opcodes",
    "evm.bytecode.sourceMap": "source_map",
    "evm.deployedBytecode.object": "bytecode_runtime",
    "evm.deployedBytecode.opcodes": "opcodes_runtime",
    "evm.deployedBytecode.sourceMap": "source_map_runtime",
    "interface": "interface",
    "ir": "ir_dict",
    "ir_runtime": "ir_runtime_dict",
    "metadata": "metadata",
    "layout": "layout",
    "userdoc": "userdoc",
}


def _parse_cli_args():
    return _parse_args(sys.argv[1:])


def _parse_args(argv):
    parser = argparse.ArgumentParser(
        description="Vyper programming language for EVM - JSON Compiler",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "input_file",
        help="JSON file to compile from. If none is given, Vyper will receive it from stdin.",
        nargs="?",
    )
    parser.add_argument(
        "--version", action="version", version=f"{vyper.__version__}+commit.{vyper.__commit__}"
    )
    parser.add_argument(
        "-o",
        help="Filename to save JSON output to. If the file exists it will be overwritten.",
        default=None,
        dest="output_file",
    )
    parser.add_argument("--pretty-json", help="Output JSON in pretty format.", action="store_true")
    parser.add_argument(
        "--traceback",
        help="Show python traceback on error instead of returning JSON",
        action="store_true",
    )

    args = parser.parse_args(argv)
    if args.input_file:
        with Path(args.input_file).open() as fh:
            input_json = fh.read()
        json_path = Path(args.input_file).resolve().as_posix()
    else:
        input_json = "".join(sys.stdin.read()).strip()
        json_path = "<stdin>"

    exc_handler = exc_handler_raises if args.traceback else exc_handler_to_dict
    output_json = json.dumps(
        compile_json(input_json, exc_handler, json_path),
        indent=2 if args.pretty_json else None,
        sort_keys=True,
        default=str,
    )

    if args.output_file is not None:
        output_path = Path(args.output_file).resolve()
        with output_path.open("w") as fh:
            fh.write(output_json)
        print(f"Results saved to {output_path}")
    else:
        print(output_json)


def exc_handler_raises(file_path: Optional[str], exception: Exception, component: str) -> None:
    if file_path:
        print(f"Unhandled exception in '{file_path}':")
    exception._exc_handler = True  # type: ignore
    raise exception


def exc_handler_to_dict(file_path: Optional[str], exception: Exception, component: str) -> dict:
    err_dict: dict = {
        "type": type(exception).__name__,
        "component": component,
        "severity": "error",
        "message": str(exception).strip('"'),
    }
    if hasattr(exception, "message"):
        err_dict.update(
            {"message": exception.message, "formattedMessage": str(exception)}  # type: ignore
        )
    if file_path is not None:
        err_dict["sourceLocation"] = {"file": file_path}
        if getattr(exception, "lineno", None) is not None:
            err_dict["sourceLocation"].update(
                {
                    "lineno": exception.lineno,  # type: ignore
                    "col_offset": getattr(exception, "col_offset", None),
                }
            )

    output_json = {"compiler": f"vyper-{vyper.__version__}", "errors": [err_dict]}
    return output_json


def get_evm_version(input_dict: dict) -> Optional[str]:
    if "settings" not in input_dict:
        return None

    # TODO: move this validation somewhere it can be reused more easily
    evm_version = input_dict["settings"].get("evmVersion")
    if evm_version is None:
        return None

    if evm_version in (
        "homestead",
        "tangerineWhistle",
        "spuriousDragon",
        "byzantium",
        "constantinople",
        "istanbul",
        "berlin",
    ):
        raise JSONError("Vyper does not support pre-london EVM versions")
    if evm_version not in EVM_VERSIONS:
        raise JSONError(f"Unknown EVM version - '{evm_version}'")

    return evm_version


def get_inputs(input_dict: dict) -> dict[PurePath, Any]:
    ret = {}
    seen = {}

    for path, value in input_dict["sources"].items():
        path = PurePath(path)
        if "urls" in value:
            raise JSONError(f"{path} - 'urls' is not a supported field, use 'content' instead")
        if "content" not in value:
            raise JSONError(f"{path} missing required field - 'content'")
        if "keccak256" in value:
            hash_ = value["keccak256"].lower().removeprefix("0x")
            if hash_ != keccak256(value["content"].encode("utf-8")).hex():
                raise JSONError(
                    f"Calculated keccak of '{path}' does not match keccak given in input JSON"
                )
        if path.stem in seen:
            raise JSONError(f"Contract namespace collision: {path}")

        # value looks like {"content": <source code>}
        # this will be interpreted by JSONInputBundle later
        ret[path] = value
        seen[path.stem] = True

    for path, value in input_dict.get("interfaces", {}).items():
        path = PurePath(path)
        if path.stem in seen:
            raise JSONError(f"Interface namespace collision: {path}")

        if isinstance(value, list):
            # backwards compatibility - straight ABI with no "abi" key.
            # (should probably just reject these)
            value = {"abi": value}

        # some validation
        if not isinstance(value, dict):
            raise JSONError(f"invalid interface (must be a dictionary):\n{json.dumps(value)}")
        if "content" in value:
            if not isinstance(value["content"], str):
                raise JSONError(f"invalid 'content' (expected string):\n{json.dumps(value)}")
        elif "abi" in value:
            if not isinstance(value["abi"], list):
                raise JSONError(f"invalid 'abi' (expected list):\n{json.dumps(value)}")
        else:
            raise JSONError(
                f"invalid interface (must contain either 'content' or 'abi'):\n{json.dumps(value)}"
            )
        if "content" in value and "abi" in value:
            raise JSONError(
                f"invalid interface (found both 'content' and 'abi'):\n{json.dumps(value)}"
            )

        ret[path] = value
        seen[path.stem] = True

    return ret


# get unique output formats for each contract, given the input_dict
# NOTE: would maybe be nice to raise on duplicated output formats
def get_output_formats(input_dict: dict) -> dict[PurePath, list[str]]:
    output_formats: dict[PurePath, list[str]] = {}
    for path, outputs in input_dict["settings"]["outputSelection"].items():
        if isinstance(outputs, dict):
            # if outputs are given in solc json format, collapse them into a single list
            outputs = OrderedSet(x for i in outputs.values() for x in i)
        else:
            outputs = OrderedSet(outputs)

        for key in [i for i in ("evm", "evm.bytecode", "evm.deployedBytecode") if i in outputs]:
            outputs.remove(key)
            outputs.update([i for i in TRANSLATE_MAP if i.startswith(key)])

        if "*" in outputs:
            outputs = TRANSLATE_MAP.values()
        else:
            try:
                outputs = [TRANSLATE_MAP[i] for i in outputs]
            except KeyError as e:
                raise JSONError(f"Invalid outputSelection - {e}")

        outputs = sorted(list(outputs))

        if path == "*":
            output_paths = [PurePath(path) for path in input_dict["sources"].keys()]
        else:
            output_paths = [PurePath(path)]
            if output_paths[0].as_posix() not in input_dict["sources"]:
                raise JSONError(f"outputSelection references unknown contract '{output_paths[0]}'")

        for output_path in output_paths:
            output_formats[output_path] = outputs

    return output_formats


def get_search_paths(input_dict: dict) -> list[PurePath]:
    ret = input_dict["settings"].get("search_paths", ".")
    return [PurePath(p) for p in ret]


def compile_from_input_dict(
    input_dict: dict, exc_handler: Callable = exc_handler_raises
) -> tuple[dict, dict]:
    if input_dict["language"] != "Vyper":
        raise JSONError(f"Invalid language '{input_dict['language']}' - Only Vyper is supported.")

    evm_version = get_evm_version(input_dict)

    optimize = input_dict["settings"].get("optimize")
    experimental_codegen = input_dict["settings"].get("experimentalCodegen", False)
    if isinstance(optimize, bool):
        # bool optimization level for backwards compatibility
        warnings.warn(
            "optimize: <bool> is deprecated! please use one of 'gas', 'codesize', 'none'.",
            stacklevel=2,
        )
        optimize = OptimizationLevel.default() if optimize else OptimizationLevel.NONE
    elif isinstance(optimize, str):
        optimize = OptimizationLevel.from_string(optimize)
    else:
        assert optimize is None

    settings = Settings(
        evm_version=evm_version, optimize=optimize, experimental_codegen=experimental_codegen
    )

    no_bytecode_metadata = not input_dict["settings"].get("bytecodeMetadata", True)

    integrity = input_dict.get("integrity")

    sources = get_inputs(input_dict)
    output_formats = get_output_formats(input_dict)
    compilation_targets = list(output_formats.keys())
    search_paths = get_search_paths(input_dict)

    input_bundle = JSONInputBundle(sources, search_paths=search_paths)

    res, warnings_dict = {}, {}
    warnings.simplefilter("always")
    for contract_path in compilation_targets:
        with warnings.catch_warnings(record=True) as caught_warnings:
            try:
                # use load_file to get a unique source_id
                file = input_bundle.load_file(contract_path)
                assert isinstance(file, FileInput)  # mypy hint
                data = vyper.compile_from_file_input(
                    file,
                    input_bundle=input_bundle,
                    output_formats=output_formats[contract_path],
                    integrity_sum=integrity,
                    settings=settings,
                    no_bytecode_metadata=no_bytecode_metadata,
                )
                assert isinstance(data, dict)
                data["source_id"] = file.source_id
            except Exception as exc:
                return exc_handler(contract_path, exc, "compiler"), {}
            res[contract_path] = data
            if caught_warnings:
                warnings_dict[contract_path] = caught_warnings

    return res, warnings_dict


# convert output of compile_input_dict to final output format
def format_to_output_dict(compiler_data: dict) -> dict:
    output_dict: dict = {"compiler": f"vyper-{vyper.__version__}", "contracts": {}, "sources": {}}
    for path, data in compiler_data.items():
        path = path.as_posix()  # Path breaks json serializability
        output_dict["sources"][path] = {"id": data["source_id"]}

        for k in ("ast_dict", "annotated_ast_dict"):
            if k in data:
                # un-translate the key
                k2 = k.removesuffix("_dict")
                output_dict["sources"][path][k2] = data[k]["ast"]

        name = PurePath(path).stem
        output_dict["contracts"][path] = {name: {}}
        output_contracts = output_dict["contracts"][path][name]

        if "ir_dict" in data:
            output_contracts["ir"] = data["ir_dict"]

        for key in ("abi", "devdoc", "interface", "metadata", "userdoc"):
            if key in data:
                output_contracts[key] = data[key]

        if "method_identifiers" in data:
            output_contracts["evm"] = {"methodIdentifiers": data["method_identifiers"]}

        evm_keys = ("bytecode", "opcodes")
        pc_maps_keys = ("source_map",)
        if any(i in data for i in evm_keys + pc_maps_keys):
            evm = output_contracts.setdefault("evm", {}).setdefault("bytecode", {})
            if "bytecode" in data:
                evm["object"] = data["bytecode"]
            if "opcodes" in data:
                evm["opcodes"] = data["opcodes"]
            if "source_map" in data:
                evm["sourceMap"] = data["source_map"]

        if any(i + "_runtime" in data for i in evm_keys + pc_maps_keys):
            evm = output_contracts.setdefault("evm", {}).setdefault("deployedBytecode", {})
            if "bytecode_runtime" in data:
                evm["object"] = data["bytecode_runtime"]
            if "opcodes_runtime" in data:
                evm["opcodes"] = data["opcodes_runtime"]
            if "source_map_runtime" in data:
                evm["sourceMap"] = data["source_map_runtime"]

    return output_dict


# https://stackoverflow.com/a/49518779
def _raise_on_duplicate_keys(ordered_pairs: list[tuple[Hashable, Any]]) -> dict:
    """
    Raise JSONError if a duplicate key exists in provided ordered list
    of pairs, otherwise return a dict.
    """
    dict_out = {}
    for key, val in ordered_pairs:
        if key in dict_out:
            raise JSONError(f"Duplicate key: {key}")
        else:
            dict_out[key] = val
    return dict_out


def compile_json(
    input_json: dict | str,
    exc_handler: Callable = exc_handler_raises,
    json_path: Optional[str] = None,
) -> dict:
    try:
        if isinstance(input_json, str):
            try:
                input_dict = json.loads(input_json, object_pairs_hook=_raise_on_duplicate_keys)
            except json.decoder.JSONDecodeError as exc:
                new_exc = JSONError(str(exc), exc.lineno, exc.colno)
                return exc_handler(json_path, new_exc, "json")
        else:
            input_dict = input_json

        try:
            compiler_data, warn_data = compile_from_input_dict(input_dict, exc_handler)
            if "errors" in compiler_data:
                return compiler_data
        except KeyError as exc:
            new_exc = JSONError(f"Input JSON missing required field: {str(exc)}")
            return exc_handler(json_path, new_exc, "json")
        except (FileNotFoundError, JSONError) as exc:
            return exc_handler(json_path, exc, "json")

        output_dict = format_to_output_dict(compiler_data)
        if warn_data:
            output_dict["errors"] = []
            for path, msg in ((k, x) for k, v in warn_data.items() for x in v):
                output_dict["errors"].append(
                    {
                        "type": msg.category.__name__,
                        "component": "compiler",
                        "severity": "warning",
                        "message": msg.message,
                        "sourceLocation": {"file": path},
                    }
                )
        return output_dict

    except Exception as exc:
        if hasattr(exc, "_exc_handler"):
            # exception was already handled by exc_handler_raises
            raise
        exc.lineno = sys.exc_info()[-1].tb_lineno  # type: ignore
        file_path = sys.exc_info()[-1].tb_frame.f_code.co_filename  # type: ignore
        return exc_handler(file_path, exc, "vyper")



import pytest

from vyper.exceptions import InstantiationException, TypeMismatch


@pytest.mark.parametrize(
    "contract",
    [
        """
foobar: int128

@external
def foo():
    self.foobar = 1
    bar: int128 = 1

    self.foobar = empty(int128)
    bar = empty(int128)

    assert self.foobar == 0
    assert bar == 0
    """,
        """
foobar: uint256

@external
def foo():
    self.foobar = 1
    bar: uint256 = 1

    self.foobar = empty(uint256)
    bar = empty(uint256)

    assert self.foobar == 0
    assert bar == 0
    """,
        """
foobar: bool

@external
def foo():
    self.foobar = True
    bar: bool = True

    self.foobar = empty(bool)
    bar = empty(bool)

    assert self.foobar == False
    assert bar == False
    """,
        """
foobar: decimal

@external
def foo():
    self.foobar = 1.0
    bar: decimal = 1.0

    self.foobar = empty(decimal)
    bar = empty(decimal)

    assert self.foobar == 0.0
    assert bar == 0.0
    """,
        """
foobar: bytes32

@external
def foo():
    self.foobar = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
    bar: bytes32 = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF

    self.foobar = empty(bytes32)
    bar = empty(bytes32)

    assert self.foobar == 0x0000000000000000000000000000000000000000000000000000000000000000
    assert bar == 0x0000000000000000000000000000000000000000000000000000000000000000
    """,
        """
foobar: address

@external
def foo():
    self.foobar = msg.sender
    bar: address = msg.sender

    self.foobar = empty(address)
    bar = empty(address)

    assert self.foobar == empty(address)
    assert bar == empty(address)
    """,
        """
@external
def foo() -> bool:
    return empty(bool)
    """,
    ],
)
def test_empty_basic_type(contract, get_contract):
    c = get_contract(contract)
    c.foo()


@pytest.mark.parametrize(
    "contract",
    [
        """
foobar: int128[3]

@external
def foo():
    self.foobar = [1, 2, 3]
    bar: int128[3] = [1, 2, 3]

    self.foobar = empty(int128[3])
    bar = empty(int128[3])

    assert self.foobar[0] == 0
    assert self.foobar[1] == 0
    assert self.foobar[2] == 0
    assert bar[0] == 0
    assert bar[1] == 0
    assert bar[2] == 0
    """,
        """
foobar: uint256[3]

@external
def foo():
    self.foobar = [1, 2, 3]
    bar: uint256[3] = [1, 2, 3]

    self.foobar = empty(uint256[3])
    bar = empty(uint256[3])

    assert self.foobar[0] == 0
    assert self.foobar[1] == 0
    assert self.foobar[2] == 0
    assert bar[0] == 0
    assert bar[1] == 0
    assert bar[2] == 0
    """,
        """
foobar: bool[3]

@external
def foo():
    self.foobar = [True, True, True]
    bar: bool[3] = [True, True, True]

    self.foobar = empty(bool[3])
    bar = empty(bool[3])

    assert self.foobar[0] == False
    assert self.foobar[1] == False
    assert self.foobar[2] == False
    assert bar[0] == False
    assert bar[1] == False
    assert bar[2] == False
    """,
        """
foobar: decimal[3]

@external
def foo():
    self.foobar = [1.0, 2.0, 3.0]
    bar: decimal[3] = [1.0, 2.0, 3.0]

    self.foobar = empty(decimal[3])
    bar = empty(decimal[3])

    assert self.foobar[0] == 0.0
    assert self.foobar[1] == 0.0
    assert self.foobar[2] == 0.0
    assert bar[0] == 0.0
    assert bar[1] == 0.0
    assert bar[2] == 0.0
    """,
        """
foobar: bytes32[3]

@external
def foo():
    self.foobar = [
        0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF,
        0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF00000000000000000000000000000000,
        0x00000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
    ]
    bar: bytes32[3] = [
        0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF,
        0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF00000000000000000000000000000000,
        0x00000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
    ]

    self.foobar = empty(bytes32[3])
    bar = empty(bytes32[3])

    assert self.foobar[0] == 0x0000000000000000000000000000000000000000000000000000000000000000
    assert self.foobar[1] == 0x0000000000000000000000000000000000000000000000000000000000000000
    assert self.foobar[2] == 0x0000000000000000000000000000000000000000000000000000000000000000
    assert bar[0] == 0x0000000000000000000000000000000000000000000000000000000000000000
    assert bar[1] == 0x0000000000000000000000000000000000000000000000000000000000000000
    assert bar[2] == 0x0000000000000000000000000000000000000000000000000000000000000000
    """,
        """
foobar: address[3]

@external
def foo():
    self.foobar = [msg.sender, msg.sender, msg.sender]
    bar: address[3] = [msg.sender, msg.sender, msg.sender]

    self.foobar = empty(address[3])
    bar = empty(address[3])

    assert self.foobar[0] == empty(address)
    assert self.foobar[1] == empty(address)
    assert self.foobar[2] == empty(address)
    assert bar[0] == empty(address)
    assert bar[1] == empty(address)
    assert bar[2] == empty(address)
    """,
    ],
)
def test_empty_basic_type_lists(contract, get_contract):
    c = get_contract(contract)
    c.foo()


@pytest.mark.parametrize(
    "contract",
    [
        """
@external
def foo() -> uint256:
    return empty(1)
    """,
        """
@external
def foo() -> decimal:
    return empty(1.0)
    """,
        """
@external
def foo() -> bytes32:
    return empty(0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF)
    """,
        """
@external
def foo() -> address:
    return empty(0xF5D4020dCA6a62bB1efFcC9212AAF3c9819E30D7)
    """,
        """
@external
def foo():
    x: uint256 = 1
    empty(x)
    """,
    ],
)
def test_clear_literals(contract, assert_compile_failed, get_contract):
    assert_compile_failed(lambda: get_contract(contract), Exception)


def test_empty_bytes(get_contract):
    code = """
foobar: Bytes[5]

@external
def foo() -> (Bytes[5], Bytes[5]):
    self.foobar = b'Hello'
    bar: Bytes[5] = b'World'

    self.foobar = empty(Bytes[5])
    bar = empty(Bytes[5])

    return (self.foobar, bar)
    """

    c = get_contract(code)
    a, b = c.foo()
    assert a == b == b""


@pytest.mark.parametrize(
    "length,value,result",
    [
        (1, "a", False),
        (1, "", True),
        (8, "helloooo", False),
        (8, "hello", False),
        (8, "", True),
        (40, "a", False),
        (40, "hellohellohellohellohellohellohellohello", False),
        (40, "", True),
    ],
)
@pytest.mark.parametrize("op", ["==", "!="])
def test_empty_string_comparison(get_contract, length, value, result, op):
    contract = f"""
@external
def foo(xs: String[{length}]) -> bool:
    return xs {op} empty(String[{length}])
    """
    c = get_contract(contract)
    if op == "==":
        assert c.foo(value) == result
    elif op == "!=":
        assert c.foo(value) != result


@pytest.mark.parametrize(
    "length,value,result",
    [
        (1, b"a", False),
        (1, b"", True),
        (8, b"helloooo", False),
        (8, b"hello", False),
        (8, b"", True),
        (40, b"a", False),
        (40, b"hellohellohellohellohellohellohellohello", False),
        (40, b"", True),
    ],
)
@pytest.mark.parametrize("op", ["==", "!="])
def test_empty_bytes_comparison(get_contract, length, value, result, op):
    contract = f"""
@external
def foo(xs: Bytes[{length}]) -> bool:
    return empty(Bytes[{length}]) {op} xs
    """
    c = get_contract(contract)
    if op == "==":
        assert c.foo(value) == result
    elif op == "!=":
        assert c.foo(value) != result


def test_empty_struct(get_contract):
    code = """
struct FOOBAR:
    a: int128
    b: uint256
    c: bool
    d: decimal
    e: bytes32
    f: address

foobar: FOOBAR

@external
def foo():
    self.foobar = FOOBAR(
        a=1,
        b=2,
        c=True,
        d=3.0,
        e=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF,
        f=msg.sender
    )
    bar: FOOBAR = FOOBAR(
        a=1,
        b=2,
        c=True,
        d=3.0,
        e=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF,
        f=msg.sender
    )

    self.foobar = empty(FOOBAR)
    bar = empty(FOOBAR)

    assert self.foobar.a == 0
    assert self.foobar.b == 0
    assert self.foobar.c == False
    assert self.foobar.d == 0.0
    assert self.foobar.e == 0x0000000000000000000000000000000000000000000000000000000000000000
    assert self.foobar.f == empty(address)

    assert bar.a == 0
    assert bar.b == 0
    assert bar.c == False
    assert bar.d == 0.0
    assert bar.e == 0x0000000000000000000000000000000000000000000000000000000000000000
    assert bar.f == empty(address)
    """

    c = get_contract(code)
    c.foo()


def test_empty_dynarray(get_contract):
    code = """
foobar: DynArray[uint256, 10]
bar: uint256

@external
def foo():
    self.bar = 1
    self.foobar = [1,2,3,4,5]
    assert len(self.foobar) == 5

    self.foobar = empty(DynArray[uint256, 10])

    assert len(self.foobar) == 0
    assert self.bar == 1
    """

    c = get_contract(code)
    c.foo()


# param empty not working yet
@pytest.mark.xfail
def test_param_empty(get_contract):
    code = """
interface Mirror:
    # reuse the contract for this test by compiling two copies of it
    def test_empty(xs: int128[111], ys: Bytes[1024], zs: Bytes[31]) -> bool: view

# a helper function which will write all over memory with random stuff
@internal
def write_junk_to_memory():
    xs: int128[1024] = empty(int128[1024])
    for i: uint256 in range(1024):
        xs[i] = -(i + 1)
@internal
def priv(xs: int128[111], ys: Bytes[1024], zs: Bytes[31]) -> bool:
    return xs[1] == 0 and ys == b'' and zs == b''

@external
def test_empty(xs: int128[111], ys: Bytes[1024], zs: Bytes[31]) -> bool:
    empty_bytes1024: Bytes[1024] = empty(Bytes[1024])
    empty_bytes31: Bytes[31] = empty(Bytes[31])
    self.write_junk_to_memory()
    # no list equality yet so test some sample values
    return xs[0] == 0 and xs[110] == 0 and ys == empty_bytes1024 and zs == empty_bytes31

@external
def pub2() -> bool:
    self.write_junk_to_memory()
    return self.priv(empty(int128[111]), empty(Bytes[1024]), empty(Bytes[31]))

@external
def pub3(x: address) -> bool:
    self.write_junk_to_memory()
    return staticcall Mirror(x).test_empty(empty(int128[111]), empty(Bytes[1024]), empty(Bytes[31]))
    """
    c = get_contract(code)
    mirror = get_contract(code)

    assert c.test_empty([0] * 111, b"", b"")
    assert c.pub2()
    assert c.pub3(mirror.address)


# return empty not working yet
@pytest.mark.xfail
def test_return_empty(get_contract):
    code = """
struct X:
    foo: int128
    bar: address
    baz: decimal
    qux: int128[1]

# a helper function which will write all over memory with random stuff
@internal
def write_junk_to_memory():
    xs: int128[1024] = empty(int128[1024])
    for i: uint256 in range(1024):
        xs[i] = -(i + 1)

@external
def a() -> uint256:
    self.write_junk_to_memory()
    return empty(uint256)

@external
def b() -> uint256[5]:
    self.write_junk_to_memory()
    return empty(uint256[5])

@external
def c() -> uint256[5][5]:
    self.write_junk_to_memory()
    return empty(uint256[5][5])

@external
def d() -> Bytes[55]:
    self.write_junk_to_memory()
    return empty(Bytes[55])

@external
def e() -> X:
    self.write_junk_to_memory()
    return empty(X)
    """
    c = get_contract(code)

    assert c.a() == 0
    assert c.b() == (0) * 5
    assert c.c() == ([0] * 5) * 5
    assert c.d() == b""
    assert c.e() == (0, "0x" + "0" * 40, 0x0, [0])


def test_map_clear(get_contract):
    code = """
big_storage: HashMap[bytes32, bytes32]

@external
def set(key: bytes32, _value: bytes32):
    self.big_storage[key] = _value

@external
def get(key: bytes32) -> bytes32:
    return self.big_storage[key]

@external
def delete(key: bytes32):
    self.big_storage[key] = empty(bytes32)
    """

    c = get_contract(code)

    key = b"test".ljust(32)
    val = b"value".ljust(32)

    assert c.get(key) == b"\x00" * 32
    c.set(key, val)
    assert c.get(key)[:5] == b"value"
    c.delete(key)
    assert c.get(key) == b"\x00" * 32


def test_map_clear_nested(get_contract):
    code = """
big_storage: HashMap[bytes32, HashMap[bytes32, bytes32]]

@external
def set(key1: bytes32, key2: bytes32, _value: bytes32):
    self.big_storage[key1][key2] = _value

@external
def get(key1: bytes32, key2: bytes32) -> bytes32:
    return self.big_storage[key1][key2]

@external
def delete(key1: bytes32, key2: bytes32):
    self.big_storage[key1][key2] = empty(bytes32)
    """

    c = get_contract(code)

    key1 = b"test1".ljust(32)
    key2 = b"test2".ljust(32)
    val = b"value".ljust(32)

    assert c.get(key1, key2) == b"\x00" * 32
    c.set(key1, key2, val)
    assert c.get(key1, key2)[:5] == b"value"
    c.delete(key1, key2)
    assert c.get(key1, key2) == b"\x00" * 32


def test_map_clear_struct(get_contract):
    code = """
struct X:
    a: int128
    b: int128

structmap: HashMap[int128, X]

@external
def set():
    self.structmap[123] = X(a=333, b=444)

@external
def get() -> (int128, int128):
    return self.structmap[123].a, self.structmap[123].b

@external
def delete():
    self.structmap[123] = empty(X)
    """

    c = get_contract(code)

    assert c.get() == (0, 0)
    c.set()
    assert c.get() == (333, 444)
    c.delete()
    assert c.get() == (0, 0)


@pytest.mark.parametrize(
    "contract",
    [
        """
@external
def foo():
    xs: uint256[10] = empty(uint256[11])
    """,
        """
@external
def bar():
    ys: Bytes[33] = empty(Bytes[32])
    """,
        """
@external
def baz():
    zs: decimal[1][1] = empty(address[1][1])
    """,
    ],
)
def test_clear_typecheck(contract, get_contract, assert_compile_failed):
    assert_compile_failed(lambda: get_contract(contract), TypeMismatch)


@pytest.mark.parametrize(
    "a,b,expected",
    [
        ("empty(Bytes[65])", "b'hello'", (b"hello", b"")),
        ("b'hello'", "empty(Bytes[33])", (b"", b"hello")),
        (
            "empty(Bytes[65])",
            "b'thirty three bytes long baby!!!!!'",
            (b"thirty three bytes long baby!!!!!", b""),
        ),
        (
            "b'thirty three bytes long baby!!!aathirty three bytes long baby!!!a'",
            "b'thirty three bytes long baby!!!aa'",
            (
                b"thirty three bytes long baby!!!aa",
                b"thirty three bytes long baby!!!aathirty three bytes long baby!!!a",
            ),
        ),
    ],
)
def test_empty_as_func_arg(get_contract, a, b, expected):
    code_a = """
@view
@external
def foo(
    a: uint256, b: Bytes[65], c: uint256, d: Bytes[33]
) -> (uint256, Bytes[33], Bytes[65], uint256):
    return a, d, b, c
    """

    code_b = f"""
interface Foo:
    def foo(
        a: uint256, b: Bytes[65], c: uint256, d: Bytes[33]
    ) -> (uint256, Bytes[33], Bytes[65], uint256): view

@view
@external
def bar(a: address) -> (uint256, Bytes[33], Bytes[65], uint256):
    return staticcall Foo(a).foo(12, {a}, 42, {b})
    """

    c1 = get_contract(code_a)
    c2 = get_contract(code_b)

    assert c2.bar(c1.address) == (12, *expected, 42)


def test_empty_array_in_event_logging(get_contract, get_logs):
    code = """
event MyLog:
    arg1: Bytes[64]
    arg2: int128[2][3]
    arg3: int128
    arg4: Bytes[64]
    arg5: uint256[3]

@external
def foo():
    log MyLog(
        b'hellohellohellohellohellohellohellohellohello',
        empty(int128[2][3]),
        314159,
        b'helphelphelphelphelphelphelphelphelphelphelp',
        empty(uint256[3])
    )
    """

    c = get_contract(code)
    c.foo()
    (log,) = get_logs(c, "MyLog")

    assert log.args.arg1 == b"hello" * 9
    assert log.args.arg2 == [[0, 0], [0, 0], [0, 0]]
    assert log.args.arg3 == 314159
    assert log.args.arg4 == b"help" * 11
    assert log.args.arg5 == [0, 0, 0]


@pytest.mark.parametrize(
    "contract",
    [
        """
@external
def test():
    a: uint256 = empty(HashMap[uint256, uint256])[0]
    """
    ],
)
def test_invalid_types(contract, get_contract, assert_compile_failed):
    assert_compile_failed(lambda: get_contract(contract), InstantiationException)



from pathlib import Path as _Path

from vyper.compiler import compile_code, compile_from_file_input

try:
    from importlib.metadata import PackageNotFoundError  # type: ignore
    from importlib.metadata import version as _version  # type: ignore
except ModuleNotFoundError:
    from importlib_metadata import PackageNotFoundError  # type: ignore
    from importlib_metadata import version as _version  # type: ignore

_commit_hash_file = _Path(__file__).parent.joinpath("vyper_git_commithash.txt")

if _commit_hash_file.exists():
    with _commit_hash_file.open() as fp:
        __commit__ = fp.read()
else:
    __commit__ = "unknown"

try:
    __version__ = _version(__name__)
except PackageNotFoundError:
    from vyper.version import version as __version__

# pep440 version with commit hash
__long_version__ = f"{__version__}+commit.{__commit__}"



import pytest

from vyper.semantics.types.user import EventT
from vyper.utils import keccak256


def keccak256_int(signature_str):
    return int.from_bytes(keccak256(signature_str.encode()), "big")


# TODO: refactor these to all use keccak256_int util instead of hardcoded constants
EVENT_ID_TESTS = [
    (
        "event MyLog: pass",
        # MyLog()
        0x375227D322F87546308335EBD7DD99C33B9D50E4DA37CE574E20471DA5E5CBE3,
    ),
    (
        """event MyLog:
    arg1: indexed(Bytes[3])""",
        # MyLog(bytes),
        0xA1BE0F4C7DDB427F96E8F62379A428E182221B02E4C3A77CC1DF3052E5D4D4BD,
    ),
    (
        """event MyLog:
    arg1: indexed(int128)
    arg2: indexed(bool)
    arg3: indexed(address)""",
        # MyLog(int128,bool,address)
        0x6D352FAAAA2891008D60036040A38A3093DD1ED2D080EEDF1E429A2F2F8D8278,
    ),
    (
        """event Potato:
    arg1: indexed(Bytes[3])
    arg2: Bytes[4]
    arg3: indexed(address)
    arg4: indexed(int128)
    """,
        # Potato(bytes,bytes,address,int128)
        0xDC82805040581BD5F35160514B94A70D2F4A1574893F1349C84472C929BD1DF0,
    ),
    (
        """event Sausage:
    arg1: int128[2]
    arg2: uint256[3]
    arg3: int128[2][2]""",
        # Sausage(int128[2],uint256[3],int128[2][2])
        0x8E4F55680D33B209630C1C6E7EFE010DB3F4CB68AB9E3D62ABB83166E41F1844,
    ),
    (
        """event MyLog:
    arg1: Bytes[4]
    arg2: indexed(String[100])
    arg3: Bytes[33]""",
        # MyLog(bytes,string,bytes)
        0x7C3BB2850EE5EAFB776ACB6E32BD31E7292FA93036CDB23D2B07AE026476F05B,
    ),
    (
        """event FooBar:
    arg1: int128
    arg2: Bytes[4]
    arg3: Bytes[39912]
    arg4: address
    arg5: address
    arg6: uint256[3]""",
        # FooBar(int128,bytes,bytes,address,address,uint256[3])
        0xD7BA39F8B5428B9F3102921170C6AE4578AA5536E05E5A86EDD36BEA67AEDD8A,
    ),
    (
        """event Nonsense:
    a: Bytes[666]
    b: int128[4][38]
    c: indexed(address[3])
    d: uint256
    e: String[1]""",
        # Nonsense(bytes,int128[4][38],address[3],uint256,string)
        0xEE1C420D3504F8A563AA99C8341A41C4C2A1A7F35665EE2F5F341CB8F451B5FC,
    ),
    (
        """event Bar:
    a: decimal[4]""",
        keccak256_int("Bar(int168[4])"),
    ),
    (
        """event Rtadr:
    a: indexed(decimal)
    b: decimal[2][5]
    c: Bytes[4]
    d: decimal[666]""",
        keccak256_int("Rtadr(int168,int168[2][5],bytes,int168[666])"),
    ),
]


@pytest.mark.parametrize("source,signature_hash", EVENT_ID_TESTS)
def test_event_id(build_node, source, signature_hash):
    node = build_node(source)
    event = EventT.from_EventDef(node)

    assert event.event_id == signature_hash



import pytest

from vyper.exceptions import (
    InvalidLiteral,
    InvalidType,
    OverflowException,
    StructureException,
    UnexpectedNodeType,
)
from vyper.semantics.types import AddressT, BoolT, BytesT, DecimalT, StringT
from vyper.semantics.types.shortcuts import BYTES32_T, INT128_T, UINT256_T
from vyper.semantics.types.utils import type_from_annotation

BASE_SOURCE = """
"Add a docstring so the target test node doesn't become one."
"""

TYPES = {
    AddressT(): "address",
    BoolT(): "bool",
    BYTES32_T: "bytes32",
    DecimalT(): "decimal",
    INT128_T: "int128",
    UINT256_T: "uint256",
    BytesT: "Bytes[1]",
    StringT: "String[1]",
}

VALID_LITERALS = {
    AddressT(): ["0x6B175474E89094C44Da98b954EedeAC495271d0F"],
    BoolT(): ["True", "False"],
    BYTES32_T: [
        "0x6b175474e89094c44da98b954eedeac495271d0f4da98b954eedeac495271d0f",
        "0x6B175474E89094C44DA98B954EEDEAC495271D0F4DA98B954EEDEAC495271D0F",
    ],
    DecimalT(): ["-1.666", "3.31337", "8008135.0", "1.2345678901"],
    INT128_T: ["-1", "0", "12", "42"],
    UINT256_T: ["0", "12", "42"],
    BytesT: ["b''", "b'this is thirty three bytes long!!'", r"b'\xbe\xef'"],
    StringT: ["''", "'hello'", "'this is thirty three chars long!!'", "'-42'"],
}


INVALID_LITERALS = [
    "None",
    "115792089237316195423570985008687907853269984665640564039457584007913129639936",
    "115792089237316195423570985008687907853269984665640564039457584007913129639935.1",
]

ALL_LITERALS = [x for v in VALID_LITERALS.values() for x in v] + INVALID_LITERALS

INVALID_NODES = [
    "self.{}",
    "[{}]",
    "({},)",
    "{}()",
    "{}(foo)",
    "{}.foo",
    "public({})",
    "-{}",
    "{}[1]",
    "self.{}[1]",
]


def do_validate_node(type_, node):
    if type_ in (BytesT, StringT):
        t = type_.from_literal(node)
        t.validate_literal(node)
    else:
        type_.validate_literal(node)


@pytest.mark.parametrize("type_", TYPES.keys())
def test_valid_literals(build_node, type_):
    sources = VALID_LITERALS[type_]
    for source in sources:
        node = build_node(source)
        do_validate_node(type_, node)


@pytest.mark.parametrize("type_", TYPES.keys())
@pytest.mark.parametrize("source", INVALID_LITERALS)
def test_invalid_literals(build_node, type_, source):
    with pytest.raises((InvalidLiteral, OverflowException, UnexpectedNodeType)):
        # build_node throws; no need to run do_validate_node
        build_node(source)


@pytest.mark.parametrize("type_,type_str", TYPES.items())
@pytest.mark.parametrize("source", INVALID_NODES + ["{}"])
def test_invalid_node(build_node, type_, type_str, source):
    source = source.format(type_str)

    with pytest.raises((InvalidLiteral, UnexpectedNodeType)):
        node = build_node(source)
        do_validate_node(type_, node)


# no literal is a valid annotation
@pytest.mark.parametrize("type_", TYPES.keys())
@pytest.mark.parametrize("source", ALL_LITERALS)
def test_from_annotation_literal(build_node, type_, source):
    with pytest.raises((InvalidType, InvalidLiteral, OverflowException)):
        node = build_node(source)
        type_from_annotation(node)


def _check_type_equals(type_, t):
    if type_ in (BytesT, StringT):
        type_ = type_()
    if not type_.compare_type(t):
        raise InvalidType(f"{type_} != {t}")


@pytest.mark.parametrize("type_,type_str", TYPES.items())
@pytest.mark.parametrize("source", INVALID_NODES)
def test_invalid_annotations(build_node, type_, type_str, source):
    source = source.format(type_str)

    with pytest.raises((StructureException, InvalidType)):
        node = build_node(source)
        t = type_from_annotation(node)
        _check_type_equals(type_, t)


@pytest.mark.parametrize("type_", TYPES.keys())
@pytest.mark.parametrize("type_str", TYPES.values())
def test_from_annotation(build_node, type_, type_str):
    if type_str == TYPES[type_]:
        node = build_node(type_str)
        t = type_from_annotation(node)
        _check_type_equals(type_, t)
    else:
        with pytest.raises(InvalidType):
            node = build_node(type_str)
            t = type_from_annotation(node)
            _check_type_equals(type_, t)



import itertools
from typing import Callable, Iterable, List

from vyper import ast as vy_ast
from vyper.exceptions import (
    CompilerPanic,
    InvalidLiteral,
    InvalidOperation,
    InvalidReference,
    InvalidType,
    OverflowException,
    StructureException,
    TypeMismatch,
    UndeclaredDefinition,
    UnknownAttribute,
    VyperException,
    ZeroDivisionException,
)
from vyper.semantics import types
from vyper.semantics.analysis.base import ExprInfo, Modifiability, ModuleInfo, VarAccess, VarInfo
from vyper.semantics.analysis.levenshtein_utils import get_levenshtein_error_suggestions
from vyper.semantics.namespace import get_namespace
from vyper.semantics.types.base import TYPE_T, VyperType
from vyper.semantics.types.bytestrings import BytesT, StringT
from vyper.semantics.types.primitives import AddressT, BoolT, BytesM_T, IntegerT
from vyper.semantics.types.subscriptable import DArrayT, SArrayT, TupleT
from vyper.utils import checksum_encode, int_to_fourbytes


def _validate_op(node, types_list, validation_fn_name):
    if not types_list:
        # TODO raise a better error here: say which types.
        raise TypeMismatch(f"Cannot perform {node.op.description} between dislike types", node)

    ret = []
    err_list = []
    for type_ in types_list:
        _validate_fn = getattr(type_, validation_fn_name)
        try:
            _validate_fn(node)
            ret.append(type_)
        except InvalidOperation as e:
            err_list.append(e)

    if ret:
        return ret

    raise err_list[0]


def uses_state(var_accesses: Iterable[VarAccess]) -> bool:
    return any(s.variable.is_state_variable() for s in var_accesses)


class _ExprAnalyser:
    """
    Node type-checker class.

    Type-check logic is implemented in `type_from_<NODE_CLASS>` methods, organized
    according to the Vyper ast node class. Calls to `get_exact_type_from_node` and
    `get_possible_types_from_node` are forwarded to this class, where the node
    class's method resolution order is examined to decide which method to call.
    """

    def __init__(self):
        self.namespace = get_namespace()

    def get_expr_info(self, node: vy_ast.VyperNode, is_callable: bool = False) -> ExprInfo:
        t = self.get_exact_type_from_node(node, include_type_exprs=is_callable)

        # if it's a Name, we have varinfo for it
        if isinstance(node, vy_ast.Name):
            info = self.namespace[node.id]

            if isinstance(info, VarInfo):
                return ExprInfo.from_varinfo(info)

            if isinstance(info, ModuleInfo):
                return ExprInfo.from_moduleinfo(info)

            if isinstance(info, VyperType):
                return ExprInfo(TYPE_T(info))

            raise CompilerPanic(f"unreachable! {info}", node)

        if isinstance(node, vy_ast.Attribute):
            # if it's an Attr, we check the parent exprinfo and
            # propagate the parent exprinfo members down into the new expr
            # note: Attribute(expr value, identifier attr)

            info = self.get_expr_info(node.value, is_callable=is_callable)
            attr = node.attr

            t = info.typ.get_member(attr, node)

            # it's a top-level variable
            if isinstance(t, VarInfo):
                return ExprInfo.from_varinfo(t, attr=attr)

            if isinstance(t, ModuleInfo):
                return ExprInfo.from_moduleinfo(t, attr=attr)

            return info.copy_with_type(t, attr=attr)

        # If it's a Subscript, propagate the subscriptable varinfo
        if isinstance(node, vy_ast.Subscript):
            info = self.get_expr_info(node.value)
            return info.copy_with_type(t)

        return ExprInfo(t)

    def get_exact_type_from_node(self, node, include_type_exprs=False):
        """
        Find exactly one type for a given node.

        Raises StructureException if a single type cannot be determined.

        Arguments
        ---------
        node : VyperNode
            The vyper AST node to find a type for.

        Returns
        -------
        Type object
        """
        types_list = self.get_possible_types_from_node(node, include_type_exprs=include_type_exprs)

        if len(types_list) > 1:
            raise StructureException("Ambiguous type", node)

        return types_list[0]

    def get_possible_types_from_node(self, node, include_type_exprs=False):
        """
        Find all possible types for a given node.
        If the node's metadata contains type information, then that type is returned.

        Arguments
        ---------
        node : VyperNode
            The vyper AST node to find a type for.

        Returns
        -------
        List
            A list of type objects
        """
        # Early termination if typedef is propagated in metadata
        if "type" in node._metadata:
            return [node._metadata["type"]]

        # this method is a perf hotspot, so we cache the result and
        # try to return it if found.
        k = f"possible_types_from_node_{include_type_exprs}"
        if k not in node._metadata:
            fn = self._find_fn(node)
            ret = fn(node)

            if not include_type_exprs:
                invalid = next((i for i in ret if isinstance(i, TYPE_T)), None)
                if invalid is not None:
                    raise InvalidReference(f"not a variable or literal: '{invalid.typedef}'", node)

            if all(isinstance(i, IntegerT) for i in ret):
                # for numeric types, sort according by number of bits descending
                # this ensures literals are cast with the largest possible type
                ret.sort(key=lambda k: (k.bits, not k.is_signed), reverse=True)

            node._metadata[k] = ret

        return node._metadata[k].copy()

    def _find_fn(self, node):
        # look for a type-check method for each class in the given class mro
        for name in [i.__name__ for i in type(node).mro()]:
            if name == "VyperNode":
                break
            fn = getattr(self, f"types_from_{name}", None)
            if fn is not None:
                return fn

        raise StructureException("Cannot determine type of this object", node)

    def types_from_Attribute(self, node):
        is_self_reference = node.get("value.id") == "self"

        # variable attribute, e.g. `foo.bar`
        t = self.get_exact_type_from_node(node.value, include_type_exprs=True)
        name = node.attr

        def _raise_invalid_reference(name, node):
            raise InvalidReference(
                f"'{name}' is not a storage variable, it should not be prepended with self", node
            )

        try:
            s = t.get_member(name, node)

            if isinstance(s, (VyperType, TYPE_T)):
                # ex. foo.bar(). bar() is a ContractFunctionT
                return [s]

            # general case. s is a VarInfo, e.g. self.foo
            if is_self_reference and (s.is_constant or s.is_immutable):
                _raise_invalid_reference(name, node)
            return [s.typ]

        except UnknownAttribute as e:
            if not is_self_reference:
                raise e from None
            if name in self.namespace:
                _raise_invalid_reference(name, node)

            hint = get_levenshtein_error_suggestions(name, t.members, 0.4)
            raise UndeclaredDefinition(
                f"Storage variable '{name}' has not been declared.", node, hint=hint
            ) from None

    def types_from_BinOp(self, node):
        # binary operation: `x + y`
        if isinstance(node.op, (vy_ast.LShift, vy_ast.RShift)):
            # ad-hoc handling for LShift and RShift, since operands
            # can be different types
            types_list = get_possible_types_from_node(node.left)
            # check rhs is unsigned integer
            validate_expected_type(node.right, IntegerT.unsigneds())
        else:
            types_list = get_common_types(node.left, node.right)

        if (
            isinstance(node.op, (vy_ast.Div, vy_ast.FloorDiv, vy_ast.Mod))
            and isinstance(node.right, vy_ast.Num)
            and not node.right.value
        ):
            raise ZeroDivisionException(f"{node.op.description} by zero", node)

        return _validate_op(node, types_list, "validate_numeric_op")

    def types_from_BoolOp(self, node):
        # boolean operation: `x and y`
        types_list = get_common_types(*node.values)
        _validate_op(node, types_list, "validate_boolean_op")
        return [BoolT()]

    def types_from_Compare(self, node):
        # comparisons, e.g. `x < y`

        # TODO fixme circular import
        from vyper.semantics.types.user import FlagT

        if isinstance(node.op, (vy_ast.In, vy_ast.NotIn)):
            # x in y
            left = self.get_possible_types_from_node(node.left)
            right = self.get_possible_types_from_node(node.right)
            if any(isinstance(t, FlagT) for t in left):
                types_list = get_common_types(node.left, node.right)
                _validate_op(node, types_list, "validate_comparator")
                return [BoolT()]

            if any(isinstance(i, SArrayT) for i in left):
                raise InvalidOperation(
                    "Left operand in membership comparison cannot be Array type", node.left
                )
            if any(not isinstance(i, (DArrayT, SArrayT)) for i in right):
                raise InvalidOperation(
                    "Right operand must be Array for membership comparison", node.right
                )
            types_list = [i for i in left if _is_type_in_list(i, [i.value_type for i in right])]
            if not types_list:
                raise TypeMismatch(
                    "Cannot perform membership comparison between dislike types", node
                )
        else:
            types_list = get_common_types(node.left, node.right)
            _validate_op(node, types_list, "validate_comparator")
        return [BoolT()]

    def types_from_ExtCall(self, node):
        call_node = node.value
        return self._find_fn(call_node)(call_node)

    def types_from_StaticCall(self, node):
        call_node = node.value
        return self._find_fn(call_node)(call_node)

    def types_from_Call(self, node):
        # function calls, e.g. `foo()` or `MyStruct()`
        var = self.get_exact_type_from_node(node.func, include_type_exprs=True)
        return_value = var.fetch_call_return(node)
        if return_value:
            if isinstance(return_value, list):
                return return_value
            return [return_value]
        raise InvalidType(f"{var} did not return a value", node)

    def types_from_Constant(self, node):
        # literal value (integer, string, etc)
        types_list = []
        for t in types.PRIMITIVE_TYPES.values():
            try:
                # clarity and perf note: will be better to construct a
                # map from node types to valid vyper types
                if not isinstance(node, t._valid_literal):
                    continue

                # special handling for bytestrings since their
                # class objects are in the type map, not the type itself
                # (worth rethinking this design at some point.)
                if t in (BytesT, StringT):
                    t = t.from_literal(node)

                # any more validation which needs to occur
                t.validate_literal(node)
                types_list.append(t)
            except VyperException:
                continue

        if types_list:
            return types_list

        # failed; prepare a good error message
        if isinstance(node, vy_ast.Num):
            raise OverflowException(
                "Numeric literal is outside of allowable range for number types", node
            )
        raise InvalidLiteral(f"Could not determine type for literal value '{node.value}'", node)

    def types_from_IfExp(self, node):
        validate_expected_type(node.test, BoolT())
        types_list = get_common_types(node.body, node.orelse)

        if not types_list:
            a = get_possible_types_from_node(node.body)[0]
            b = get_possible_types_from_node(node.orelse)[0]
            raise TypeMismatch(f"Dislike types: {a} and {b}", node)

        return types_list

    def types_from_List(self, node):
        # literal array
        if _is_empty_list(node):
            ret = []

            if len(node.elements) > 0:
                # empty nested list literals `[[], []]`
                subtypes = self.get_possible_types_from_node(node.elements[0])
            else:
                # empty list literal `[]`
                # subtype can be anything
                subtypes = types.PRIMITIVE_TYPES.values()

            for t in subtypes:
                # 1 is minimum possible length for dynarray,
                # can be assigned to anything
                if isinstance(t, VyperType):
                    ret.append(DArrayT(t, 1))
                elif isinstance(t, type) and issubclass(t, VyperType):
                    # for typeclasses like bytestrings, use a generic type acceptor
                    ret.append(DArrayT(t.any(), 1))
                else:
                    raise CompilerPanic(f"busted type {t}", node)
            return ret

        types_list = get_common_types(*node.elements)

        if len(types_list) > 0:
            count = len(node.elements)
            ret = []
            ret.extend([SArrayT(t, count) for t in types_list])
            ret.extend([DArrayT(t, count) for t in types_list])
            return ret
        raise InvalidLiteral("Array contains multiple, incompatible types", node)

    def types_from_Name(self, node):
        # variable name, e.g. `foo`
        name = node.id
        if (
            name not in self.namespace
            and "self" in self.namespace
            and name in self.namespace["self"].typ.members
        ):
            raise InvalidReference(
                f"'{name}' is a storage variable, access it as self.{name}", node
            )
        try:
            t = self.namespace[node.id]
            # when this is a type, we want to lower it
            if isinstance(t, VyperType):
                # TYPE_T is used to handle cases where a type can occur in call or
                # attribute conditions, like Flag.foo or MyStruct({...})
                return [TYPE_T(t)]

            return [t.typ]

        except VyperException as exc:
            raise exc.with_annotation(node) from None

    def types_from_Subscript(self, node):
        # index access, e.g. `foo[1]`
        if isinstance(node.value, (vy_ast.List, vy_ast.Subscript)):
            types_list = self.get_possible_types_from_node(node.value)
            ret = []
            for t in types_list:
                t.validate_index_type(node.slice)
                ret.append(t.get_subscripted_type(node.slice))
            return ret

        t = self.get_exact_type_from_node(node.value)
        t.validate_index_type(node.slice)
        return [t.get_subscripted_type(node.slice)]

    def types_from_Tuple(self, node):
        types_list = [self.get_exact_type_from_node(i) for i in node.elements]
        return [TupleT(types_list)]

    def types_from_UnaryOp(self, node):
        # unary operation: `-foo`
        types_list = self.get_possible_types_from_node(node.operand)
        return _validate_op(node, types_list, "validate_numeric_op")


def _is_empty_list(node):
    # Checks if a node is a `List` node with an empty list for `elements`,
    # including any nested `List` nodes. ex. `[]` or `[[]]` will return True,
    # [1] will return False.
    if not isinstance(node, vy_ast.List):
        return False

    if not node.elements:
        return True
    return all(_is_empty_list(t) for t in node.elements)


def _is_type_in_list(obj, types_list):
    # check if a type object is in a list of types
    return any(i.compare_type(obj) for i in types_list)


# NOTE: dead fn
def _filter(type_, fn_name, node):
    # filter function used when evaluating boolean ops and comparators
    try:
        getattr(type_, fn_name)(node)
        return True
    except InvalidOperation:
        return False


def get_possible_types_from_node(node):
    """
    Return a list of possible types for the given node.

    Raises if no possible types can be found.

    Arguments
    ---------
    node : VyperNode
        A vyper ast node.

    Returns
    -------
    List
        List of one or more BaseType objects.
    """
    return _ExprAnalyser().get_possible_types_from_node(node, include_type_exprs=True)


def get_exact_type_from_node(node):
    """
    Return exactly one type for a given node.

    Raises if there is more than one possible type.

    Arguments
    ---------
    node : VyperNode
        A vyper ast node.

    Returns
    -------
    BaseType
        Type object.
    """
    return _ExprAnalyser().get_exact_type_from_node(node, include_type_exprs=True)


def get_expr_info(node: vy_ast.ExprNode, is_callable: bool = False) -> ExprInfo:
    if node._expr_info is None:
        node._expr_info = _ExprAnalyser().get_expr_info(node, is_callable)
    return node._expr_info


def get_common_types(*nodes: vy_ast.VyperNode, filter_fn: Callable = None) -> List:
    # this function is a performance hotspot
    """
    Return a list of common possible types between one or more nodes.

    Arguments
    ---------
    *nodes : VyperNode
        Vyper ast nodes.
    filter_fn : Callable, optional
        If given, results are filtered by this function prior to returning.

    Returns
    -------
    list
        List of zero or more `BaseType` objects.
    """
    common_types = _ExprAnalyser().get_possible_types_from_node(nodes[0])

    for item in nodes[1:]:
        new_types = _ExprAnalyser().get_possible_types_from_node(item)

        tmp = []
        for c in common_types:
            for t in new_types:
                if t.compare_type(c) or c.compare_type(t):
                    tmp.append(c)
                    break

        common_types = tmp

    if filter_fn is not None:
        common_types = [i for i in common_types if filter_fn(i)]

    return common_types


# TODO push this into `ArrayT.validate_literal()`
def _validate_literal_array(node, expected):
    # validate that every item within an array has the same type
    if isinstance(expected, SArrayT):
        if len(node.elements) != expected.length:
            return False
    if isinstance(expected, DArrayT):
        if len(node.elements) > expected.length:
            return False

    for item in node.elements:
        try:
            validate_expected_type(item, expected.value_type)
        except (InvalidType, TypeMismatch):
            return False

    return True


def validate_expected_type(node, expected_type):
    """
    Validate that the given node matches the expected type(s)

    Raises if the node does not match one of the expected types.

    Arguments
    ---------
    node : VyperNode
        Vyper ast node.
    expected_type : Tuple | BaseType
        A type object, or tuple of type objects

    Returns
    -------
    None
    """
    if not isinstance(expected_type, tuple):
        expected_type = (expected_type,)

    if isinstance(node, vy_ast.Tuple):
        possible_tuple_types = [t for t in expected_type if isinstance(t, TupleT)]
        for t in possible_tuple_types:
            if len(t.member_types) != len(node.elements):
                continue
            for item_ast, item_type in zip(node.elements, t.member_types):
                try:
                    validate_expected_type(item_ast, item_type)
                    return
                except VyperException:
                    pass
        else:
            # fail block
            pass

    given_types = _ExprAnalyser().get_possible_types_from_node(node)

    if isinstance(node, vy_ast.List):
        # special case - for literal arrays we individually validate each item
        for expected in expected_type:
            if not isinstance(expected, (DArrayT, SArrayT)):
                continue
            if _validate_literal_array(node, expected):
                return
    else:
        for given, expected in itertools.product(given_types, expected_type):
            if expected.compare_type(given):
                return

    # validation failed, prepare a meaningful error message
    if len(expected_type) > 1:
        expected_str = f"one of {', '.join(str(i) for i in expected_type)}"
    else:
        expected_str = expected_type[0]

    if len(given_types) == 1 and getattr(given_types[0], "_is_callable", False):
        raise StructureException(
            f"{given_types[0]} cannot be referenced directly, it must be called", node
        )

    if not isinstance(node, (vy_ast.List, vy_ast.Tuple)) and node.get_descendants(
        vy_ast.Name, include_self=True
    ):
        given = given_types[0]
        raise TypeMismatch(f"Given reference has type {given}, expected {expected_str}", node)
    else:
        if len(given_types) == 1:
            given_str = str(given_types[0])
        else:
            types_str = sorted(str(i) for i in given_types)
            given_str = f"{', '.join(types_str[:1])} or {types_str[-1]}"

        suggestion_str = ""
        if expected_type[0] == AddressT() and given_types[0] == BytesM_T(20):
            suggestion_str = f" Did you mean {checksum_encode(node.value)}?"

        raise TypeMismatch(
            f"Expected {expected_str} but literal can only be cast as {given_str}.{suggestion_str}",
            node,
        )


def validate_unique_method_ids(functions: List) -> None:
    """
    Check for collisions between the 4byte function selectors
    of each function within a contract.

    Arguments
    ---------
    functions : List[ContractFunctionT]
        A list of ContractFunctionT objects.
    """
    method_ids = [x for i in functions for x in i.method_ids.values()]
    seen = set()
    for method_id in method_ids:
        if method_id in seen:
            collision_str = ", ".join(
                x for i in functions for x in i.method_ids.keys() if i.method_ids[x] == method_id
            )
            collision_hex = int_to_fourbytes(method_id).hex()
            raise StructureException(
                f"Methods produce colliding method ID `0x{collision_hex}`: {collision_str}"
            )
        seen.add(method_id)


def check_modifiability(node: vy_ast.ExprNode, modifiability: Modifiability) -> bool:
    """
    Check if the given node is not more modifiable than the given modifiability.
    """
    if node.is_literal_value or node.has_folded_value:
        return True

    if isinstance(node, (vy_ast.BinOp, vy_ast.Compare)):
        return all(check_modifiability(i, modifiability) for i in (node.left, node.right))

    if isinstance(node, vy_ast.BoolOp):
        return all(check_modifiability(i, modifiability) for i in node.values)

    if isinstance(node, vy_ast.UnaryOp):
        return check_modifiability(node.operand, modifiability)

    if isinstance(node, (vy_ast.Tuple, vy_ast.List)):
        return all(check_modifiability(item, modifiability) for item in node.elements)

    if isinstance(node, vy_ast.Call):
        call_type = get_exact_type_from_node(node.func)

        # structs and interfaces
        if hasattr(call_type, "check_modifiability_for_call"):
            return call_type.check_modifiability_for_call(node, modifiability)

    info = get_expr_info(node)
    return info.modifiability <= modifiability
